import cocos.device
import cocos.numerics as cn
import cocos.numerics.linalg as linalg
import numpy as np
import pytest

test_data = [(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 20]],
                       dtype=np.float64),
             np.array([[100, 1000], [200, 2000], [300, 3000]],
                      dtype=np.float64))]


@pytest.mark.parametrize("A_numpy, b_numpy", test_data)
def test_solve(A_numpy, b_numpy):
    cocos.device.init()
    if cocos.device.is_dbl_supported():
        # using numpy
        # print(A)
        solution = np.linalg.solve(A_numpy, b_numpy)

        # using Cocos
        A_cocos = cn.array(A_numpy)
        b_cocos = cn.array(b_numpy)

        solution_cocos = cn.linalg.solve(A_cocos, b_cocos)

        # conduct tests
        assert np.allclose(solution, solution_cocos)


import pytest
import typing as tp

from cocos.multi_processing.utilities import (
    generate_slices_with_batch_size,
    generate_slices_with_number_of_batches
)


@pytest.mark.parametrize('n, batch_size, result', [(10, 2, ((0, 2), (2, 4), (4, 6), (6, 8), (8, 10))),
                                                   (10, 3, ((0, 3), (3, 6), (6, 9), (9, 10))),
                                                   (10, 4, ((0, 4), (4, 8), (8, 10)))])
def test_generate_slices_with_batch_size(n: int, batch_size: int, result: tp.Tuple[tp.Tuple[int, int]]):
    assert result == generate_slices_with_batch_size(n=n, batch_size=batch_size)


@pytest.mark.parametrize('n, number_of_batches, result', [(10, 2, ((0, 5), (5, 10))),
                                                          (10, 3, ((0, 4), (4, 8), (8, 10))),
                                                          (10, 4, ((0, 3), (3, 6), (6, 9), (9, 10)))])
def test_generate_slices_with_batch_size(n: int, number_of_batches: int, result: tp.Tuple[tp.Tuple[int, int]]):
    assert result == generate_slices_with_number_of_batches(n=n, number_of_batches=number_of_batches)



import collections
import typing as tp

import numpy as np
import sympy as sym

from cocos import numerics as cn
from cocos.numerics.data_types import NumericArray, NumericArrayOrScalar
from cocos.utilities import check_and_make_sequence
from cocos.numerics.numerical_package_selector import \
    select_num_pack, \
    get_gpu_and_num_pack_by_dtype_from_iterable
from cocos.symbolic.translations import COCOS_TRANSLATIONS
from cocos.symbolic.utilities import find_length_of_state_vectors


DUMMY_TIME_SYMBOL = sym.Symbol('t')

################################################################################
# Works for both Matrices and Arrays
################################################################################
def _compute_replacement_functions(
        state_vectors: tp.List[NumericArrayOrScalar],
        number_of_state_variables: int,
        t: tp.Optional[float] = None,
        replacement_functions: tp.Optional[tp.Tuple[tp.Callable, ...]] = None) \
        -> tp.Tuple[tp.List, int]:
    """
    If an expression has been optimized with common subexpression elimination,
    one first needs to compute the replacement values (which are additional
    arguments to the original function) using so-called replacement functions.
    This is done in an incremental fashion, i.e. the replacement value resulting
    from a given replacement function may be used as an argument of the next
    replacement function.

    This function evaluates these replacement functions with the original
    arguments and returns the replacement values appended to the original
    arguments along with the number of replications.

    Args:
        replacement_functions:
            A sequence of functions that compute the replacement values from the
            original arguments.

        state_vectors: a list of numerical vectors (NumPy or Cocos)
        t: time parameter
        gpu: whether or not to evaluate the function on the gpu
        number_of_state_variables:
            the number of variables (original arguments) in the function

    Returns:
        a tuple of arguments with replacement values appended and the number of
        replications

    """

    # R = max([state_argument.size
    #          for state_argument
    #          in state_vectors
    #          if isinstance(state_argument, (np.ndarray, cn.ndarray))] + [1])

    R = find_length_of_state_vectors(state_vectors)

    if number_of_state_variables != len(state_vectors):
        raise ValueError(f"The number of state vectors({len(state_vectors)}) "
                         f"does not match the number of variables in the "
                         f"function({number_of_state_variables}).")

    arguments = [t] + state_vectors
    if replacement_functions is not None:
        for replacement_function in replacement_functions:
            # Note that the arguments vector grows with each iteration. That
            # means computed common subexpressions from previous iterations are
            # used to evaluate subsequent expressions.

            arguments.append(replacement_function(*arguments))

    return arguments, R


################################################################################
# General Arrays
################################################################################
def lambdify_array_with_modules(
        symbols: tp.Tuple[sym.Symbol, ...],
        array_expression: tp.Union[sym.Array, sym.MatrixBase],
        symbolic_time_function_name_to_numeric_time_function_map:
        tp.Dict[str, tp.Callable],
        modules: tp.Tuple[str, ...]) \
        -> tp.Tuple[tp.Callable, ...]:
    """
    This function takes a SymPy array, unravals the index in Fortran order and
    lambdifies each element separately.

    Args:
        symbols: a tuple of SymPy symbols that are arguments to the function
        array_expression: a SymPy matrix or array
        symbolic_time_function_name_to_numeric_time_function_map: 
            a dictionary mapping the names of functions of time to Python 
            functions
        modules: modules that should be included

    Returns: a tuple of functions of the specified argument representing the
             elements of the symbolic array in Fortran order

    """

    if not isinstance(modules, list):
        modules = list(modules)

    if not (isinstance(array_expression, sym.Array) or
            isinstance(array_expression, sym.MatrixBase)):
        raise TypeError(f"symbolic_matrix_expression must be an instance of "
                        f"sympy.ArrayBase or sympy.MatrixBase "
                        f"but is of type {type(array_expression)}")

    n = (np.prod(array_expression.shape)).item()
    result = []
    for i in range(n):
        index = tuple(np.unravel_index(i, array_expression.shape, order='F'))
        result.append(sym.lambdify(args=symbols,
                                   expr=array_expression[index],
                                   modules
                                   =[symbolic_time_function_name_to_numeric_time_function_map] +
                                    modules))

    return tuple(result)


def lambdify_array(
        symbols: tp.Tuple[sym.Symbol, ...],
        array_expression: tp.Union[sym.Array, sym.MatrixBase],
        symbolic_time_function_name_to_numeric_time_function_map:
        tp.Optional[tp.Dict[str, tp.Callable]] = None) \
        -> tp.Tuple[tp.Tuple[tp.Callable, ...], tp.Tuple[tp.Callable, ...]]:

    if symbolic_time_function_name_to_numeric_time_function_map is None:
        symbolic_time_function_name_to_numeric_time_function_map = dict()

    functions_cpu = \
        lambdify_array_with_modules(
            symbols=symbols,
            array_expression=array_expression,
            symbolic_time_function_name_to_numeric_time_function_map
            =symbolic_time_function_name_to_numeric_time_function_map,
            modules=('numpy', ))

    functions_gpu = \
        lambdify_array_with_modules(
            symbols=symbols,
            array_expression=array_expression,
            symbolic_time_function_name_to_numeric_time_function_map
            =symbolic_time_function_name_to_numeric_time_function_map,
            modules=(COCOS_TRANSLATIONS, ))

    return functions_cpu, functions_gpu


def _compute_result_internal(R: int,
                             dimensions: tp.Tuple[int, ...],
                             arguments: tp.List[NumericArrayOrScalar],
                             functions_cpu: tp.Tuple[tp.Callable, ...],
                             functions_gpu: tp.Tuple[tp.Callable, ...],
                             pre_attach: bool,
                             gpu: bool,
                             dtype: np.generic) \
        -> NumericArray:
    """
    This function evaluates a sequence of functions with given positional
    arguments. Each argument is either a scalar or R-dimensional. The result
    Args:
        R: the number of replications (this is the length each of the input
           vectors if they are not scalars)

        dimensions:
            The shape of the output (excluding the R replications which are
            either attached to the beginning (if pre_attach is True) or to the
            end.

        arguments:
            a list of arguments to be passed to the functions to be evaluated

        functions_cpu: The sequence of functions to be evaluated on the CPU.
        functions_gpu: The sequence of functions to be evaluated on the GPU.
        pre_attach:
            whether to put the replications in the first or the last axis
        gpu: whether to evaluate on the cpu or the gpu
        dtype: the dtype of the output array

    Returns:

    """
    num_pack = select_num_pack(gpu)
    if gpu:
        functions = functions_gpu
    else:
        functions = functions_cpu

    n = (np.prod(dimensions)).item()

    if pre_attach:
        dimension_index = tuple([R] + list(dimensions))
    else:
        dimension_index = tuple(list(dimensions) + [R])

    result = num_pack.zeros(dimension_index, dtype=dtype)

    for i in range(n):
        index = tuple(np.unravel_index(i, dimensions, order='F'))

        function = functions[i]

        if pre_attach:
            location_index = tuple([slice(None)] + list(index))
        else:
            location_index = tuple(list(index) + [slice(None)])

        evaluated_function = function(*arguments)
        result[location_index] = evaluated_function

    return result


################################################################################
# Sequences of Lambdified Array, Matrix, and Vector-Expressions
################################################################################
class LambdifiedArrayExpressions:
    def __init__(
            self,
            symbolic_array_expressions:
            tp.Tuple[tp.Union[sym.Array, sym.MatrixBase], ...],
            argument_symbols: tp.Tuple[sym.Symbol],
            time_symbol: tp.Optional[sym.Symbol] = None,
            symbolic_time_function_name_to_numeric_time_function_map:
            tp.Optional[tp.Dict[str, tp.Callable]] = None,
            squeeze_column_vectors: tp.Optional[tp.Tuple[bool, ...]] = None,
            perform_cse: bool = True,
            lazy_initialization: bool = False,
            pre_attach: tp.Optional[tp.Tuple[bool, ...]] = None,
            dtype: np.generic = np.float32):

        if time_symbol is None:
            if symbolic_time_function_name_to_numeric_time_function_map is not None and \
                    len(symbolic_time_function_name_to_numeric_time_function_map) != 0:
                raise ValueError(
                        'Argument time_symbol must not be none if '
                        'symbolic_time_function_name_to_numeric_time_function_map '
                        'is provided.')

            if any([DUMMY_TIME_SYMBOL
                    in symbolic_array_expression.free_symbols
                    for symbolic_array_expression
                    in symbolic_array_expressions]):
                raise ValueError('Argument time_symbol must be provided '
                                 'explicity if expression contains '
                                 'DUMMY_TIME_SYMBOL.')

            time_symbol = DUMMY_TIME_SYMBOL
            self._time_symbol_provided = False
        else:
            self._time_symbol_provided = True

        if symbolic_time_function_name_to_numeric_time_function_map is None:
            symbolic_time_function_name_to_numeric_time_function_map = dict()

        symbolic_array_expressions \
            = tuple(check_and_make_sequence(symbolic_array_expressions,
                                            sym.MatrixBase))

        # print(f'{symbolic_array_expressions=}\n')
        symbolic_array_expressions = \
                tuple(sym.ImmutableMatrix(symbolic_array_expression)
                      if isinstance(symbolic_array_expression, sym.MatrixBase)
                      else sym.ImmutableDenseNDimArray(symbolic_array_expression)
                      for symbolic_array_expression
                      in symbolic_array_expressions)

        if pre_attach is None:
            pre_attach = len(symbolic_array_expressions) * [True]
        pre_attach = check_and_make_sequence(pre_attach, bool)

        if squeeze_column_vectors is None:
            squeeze_column_vectors = len(symbolic_array_expressions) * [False]
        squeeze_column_vectors = check_and_make_sequence(squeeze_column_vectors,
                                                         bool)

        if not (len(symbolic_array_expressions) == len(pre_attach) and
                len(symbolic_array_expressions) == len(squeeze_column_vectors)):
            raise ValueError("symbolic_matrix_expressions, pre_attach, and "
                             "squeeze_column_vectors must have the same number "
                             "of elements")

        self._argument_symbols = argument_symbols
        self._time_symbol = time_symbol
        self._symbolic_array_expressions = symbolic_array_expressions
        self._symbolic_time_function_name_to_numeric_time_function_map = \
            symbolic_time_function_name_to_numeric_time_function_map
        self._squeeze_column_vectors = squeeze_column_vectors
        self._symbols = \
            tuple([time_symbol] +
                   list(argument_symbols))

        self._perform_cse = perform_cse
        self._pre_attach = pre_attach
        self._dtype = dtype

        self._shapes \
            = tuple([symbolic_array_expression.shape
                     for symbolic_array_expression
                     in symbolic_array_expressions])

        if lazy_initialization:
            self._functions_cpu = None
            self._functions_gpu = None
        else:
            self._perform_initialization()

    def _perform_initialization(self):
        self._functions_cpu = []
        self._functions_gpu = []
        if self._perform_cse:
            repl, redu = sym.cse(self._symbolic_array_expressions,
                                 optimizations='basic')

            replacement_functions_cpu = []
            replacement_functions_gpu = []
            syms = [self._time_symbol] + list(self._argument_symbols)
            for i, v in enumerate(repl):
                replacement_functions_cpu.append(
                    sym.lambdify(args=syms,
                                 expr=v[1],
                                 modules
                                 =[self._symbolic_time_function_name_to_numeric_time_function_map,
                                   'numpy']))

                replacement_functions_gpu.append(
                    sym.lambdify(
                        args=syms,
                        expr=v[1],
                        modules
                        =[self._symbolic_time_function_name_to_numeric_time_function_map,
                          COCOS_TRANSLATIONS]))

                syms.append(v[0])

            self._replacement_functions_cpu = tuple(replacement_functions_cpu)
            self._replacement_functions_gpu = tuple(replacement_functions_gpu)

            for symbolic_matrix_expression in redu:
                functions_cpu, functions_gpu \
                    = lambdify_array(
                        symbols=tuple(syms),
                        array_expression=symbolic_matrix_expression,
                        symbolic_time_function_name_to_numeric_time_function_map
                        =self._symbolic_time_function_name_to_numeric_time_function_map)

                self._functions_cpu.append(functions_cpu)
                self._functions_gpu.append(functions_gpu)
        else:
            for symbolic_matrix_expression in self._symbolic_array_expressions:
                functions_cpu, functions_gpu \
                    = lambdify_array(
                        symbols=self._symbols,
                        array_expression=symbolic_matrix_expression,
                        symbolic_time_function_name_to_numeric_time_function_map
                        =self._symbolic_time_function_name_to_numeric_time_function_map)

                self._functions_cpu.append(functions_cpu)
                self._functions_gpu.append(functions_gpu)

    @property
    def shapes(self) -> tp.Tuple[tp.Tuple[int, ...], ...]:
        return self._shapes

    def is_column_vector(self, i: int) -> bool:
        shape = self._symbolic_array_expressions[i].shape
        return len(shape) == 1 \
               or all([dim == 1
                       for dim
                       in shape[1:]])

    def is_row_vector(self, i: int) -> bool:
        shape = self._symbolic_array_expressions[i].shape
        if len(shape < 2):
            return False
        else:
            return shape[0] == 1 and \
                   (len(shape) == 2 or
                    all([dim == 1
                         for dim
                         in shape[2:]]))

    @property
    def argument_symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._argument_symbols

    @property
    def number_of_state_variables(self) -> int:
        return len(self._argument_symbols)

    @property
    def time_symbol(self) -> sym.Symbol:
        return self._time_symbol

    @property
    def symbolic_array_expressions(self) -> tp.Tuple[sym.ImmutableMatrix, ...]:
        return self._symbolic_array_expressions

    @property
    def symbolic_time_function_name_to_numeric_time_function_map(self) \
            -> tp.Dict[str, tp.Callable]:
        return self._symbolic_time_function_name_to_numeric_time_function_map

    @property
    def symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._symbols

    @property
    def number_of_variables(self) -> int:
        return len(self._symbols)

    @property
    def dtype(self) -> np.generic:
        return self._dtype

    def evaluate_with_list_of_state_vectors(
            self,
            list_of_state_vectors: tp.List[NumericArrayOrScalar],
            t: tp.Optional[float] = None,
            gpu: tp.Optional[bool] = None) \
            -> tp.Tuple[NumericArray, ...]:
        """
        This function evaluates the array expressions at arguments that are
        vectors (arrays with a single axis) or scalars and returns the resulting
        arrays as a tuple.

        Args:
            list_of_state_vectors:
                A list of one-dimensional numeric arrays. The length of this
                list must match the number of arguments to the array-valued
                functions.

            t: time parameter
            gpu: whether to evaluate the array expressions on the GPU

        Returns: a tuple of arrays corresponding to the array expressions
                 evaluated at the parameter vectors

        """
        if t is None:
            if self._time_symbol_provided:
                raise ValueError("Argument 't' must be present if 'time_symbol'"
                                 " was explicitly provided during "
                                 "construction.'")

            if (self._symbolic_time_function_name_to_numeric_time_function_map is not None and
                    len(self._symbolic_time_function_name_to_numeric_time_function_map) > 0):
                raise ValueError(
                        "Argument 't' must be provided if "
                        "symbolic_time_function_name_to_numeric_time_function_map "
                        "is provided.")

        if self._functions_cpu is None or self._functions_gpu is None:
            self._perform_initialization()

        if gpu is None:
            # determine whether to run on cpu or gpu based on the input types
            gpu, _ \
                = get_gpu_and_num_pack_by_dtype_from_iterable(
                list_of_state_vectors)

        if gpu:
            replacement_functions = self._replacement_functions_gpu
        else:
            replacement_functions = self._replacement_functions_cpu

        arguments, R \
            = _compute_replacement_functions(
                state_vectors=list_of_state_vectors,
                number_of_state_variables=self.number_of_state_variables,
                t=t,
                replacement_functions=replacement_functions)

        results = []
        for i, (shape,
                pre_attach,
                functions_cpu,
                functions_gpu,
                squeeze_column_vector) in \
                enumerate(zip(self.shapes,
                              self._pre_attach,
                              self._functions_cpu,
                              self._functions_gpu,
                              self._squeeze_column_vectors)):

            is_column_vector = self.is_column_vector(i)

            if is_column_vector and squeeze_column_vector:
                dimensions = (shape[0], )
            else:
                dimensions = shape

            result = _compute_result_internal(R=R,
                                              dimensions=dimensions,
                                              arguments=arguments,
                                              functions_cpu=functions_cpu,
                                              functions_gpu=functions_gpu,
                                              pre_attach=pre_attach,
                                              gpu=gpu,
                                              dtype=self.dtype)
            results.append(result)

        return tuple(results)

    def evaluate_with_dictionary(
            self,
            symbolic_to_numeric_parameter_map: tp.Dict[sym.Symbol,
                                                       NumericArrayOrScalar],
            t: tp.Optional[float] = None,
            gpu: tp.Optional[bool] = None) \
            -> tp.Tuple[NumericArray, ...]:
        """
        This function evaluates the array expressions at arguments that are
        dictionaries mapping symbolic parameters to one-dimensional numeric
        arrays or scalars and returns the resulting arrays as a tuple.

        Args:
            symbolic_to_numeric_parameter_map:
                A dictionary mapping symbolic parameters to numeric arrays or
                scalars.

            t: time parameter
            gpu: whether to evaluate the array expressions on the GPU

        Returns: a tuple of arrays corresponding to the array expressions
                 evaluated at the parameter vectors

        """
        list_of_state_vectors = \
            [symbolic_to_numeric_parameter_map[symbol]
             for symbol
             in self.argument_symbols]

        return self.evaluate_with_list_of_state_vectors(
                        list_of_state_vectors=list_of_state_vectors,
                        t=t,
                        gpu=gpu)

    def evaluate_with_kwargs(self,
                             t: tp.Optional[float] = None,
                             gpu: tp.Optional[bool] = None,
                             **kwargs) \
            -> tp.Tuple[NumericArray, ...]:
        """
        This function evaluates the array expressions using the symbol names
        given by keyword arguments.

        Args:
            t: time parameter
            gpu: whether to evaluate the array expressions on the GPU
            kwargs: a dictionary mapping parameter names to numeric arrays or
                    scalars

        Returns: a tuple of arrays corresponding to the array expressions
                 evaluated at the parameter vectors

        """
        list_of_state_vectors = \
            [kwargs[symbol.name]
             for symbol
             in self.argument_symbols]

        return self.evaluate_with_list_of_state_vectors(
                        list_of_state_vectors=list_of_state_vectors,
                        t=t,
                        gpu=gpu)

    def evaluate(self,
                 state_matrices: tp.Tuple[NumericArrayOrScalar, ...],
                 t: tp.Optional[float] = None,
                 gpu: tp.Optional[bool] = None) \
            -> tp.Tuple[NumericArray, ...]:
        """
        This function evaluates the array expressions at arguments that are
        represented either as vectors or as horizontally concatenated vectors
        (i.e. arrays with 2 axes). It decomposes these matrices into column
        vectors and evaluates the array functions with this list of column
        vectors.

        The function returns a tuple of result arrays (one for each array
        function).

        Args:
            state_matrices: a tuple of matrices, vectors, or scalars
            t: time parameter
            gpu: whether to evaluate the array expressions on the GPU

        Returns: a tuple of arrays corresponding to the array expressions
                 evaluated at the parameter vectors

        """
        if not isinstance(state_matrices, collections.abc.Sequence):
            if isinstance(state_matrices, (np.ndarray, cn.ndarray)):
                state_matrices = [state_matrices]
            else:
                raise TypeError("state_variables must be of type "
                                "Sequence[Union[numpy.ndarray, cocos.ndarray]]")

        list_of_state_vectors = []
        # If a state matrix has more than one axis, it is separated it into a
        # list of vectors and appended to the list of vector arguments.
        for state_matrix in state_matrices:
            if state_matrix.ndim > 1:
                for i in range(state_matrix.shape[1]):
                    list_of_state_vectors.append((state_matrix[:, i]))
            else:
                list_of_state_vectors.append(state_matrix)

        return self.evaluate_with_list_of_state_vectors(
                        list_of_state_vectors=list_of_state_vectors,
                        t=t,
                        gpu=gpu)


class LambdifiedMatrixExpressions(LambdifiedArrayExpressions):
    def __init__(
         self,
         symbolic_matrix_expressions: tp.Tuple[sym.MatrixBase, ...],
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: tp.Optional[sym.Symbol] = None,
         symbolic_time_function_name_to_numeric_time_function_map:
         tp.Optional[tp.Dict[str, tp.Callable]] = None,
         squeeze_column_vectors: tp.Optional[tp.Tuple[bool, ...]] = None,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: tp.Optional[tp.Tuple[bool, ...]] = None,
         dtype: np.generic = np.float32):

        if symbolic_time_function_name_to_numeric_time_function_map is None:
            symbolic_time_function_name_to_numeric_time_function_map = dict()

        super().__init__(
                    argument_symbols=argument_symbols,
                    time_symbol=time_symbol,
                    symbolic_array_expressions=symbolic_matrix_expressions,
                    symbolic_time_function_name_to_numeric_time_function_map
                    =symbolic_time_function_name_to_numeric_time_function_map,
                    squeeze_column_vectors=squeeze_column_vectors,
                    perform_cse=perform_cse,
                    lazy_initialization=lazy_initialization,
                    pre_attach=pre_attach,
                    dtype=dtype)

        rows = []
        cols = []

        for symbolic_matrix_expression in symbolic_matrix_expressions:
            r, c = symbolic_matrix_expression.shape
            rows.append(r)
            cols.append(c)

        self._rows = tuple(rows)
        self._cols = tuple(cols)

    @property
    def rows(self) -> tp.Tuple[int, ...]:
        return self._rows

    @property
    def cols(self) -> tp.Tuple[int, ...]:
        return self._cols

    @property
    def symbolic_matrix_expressions(self) -> tp.Tuple[sym.ImmutableMatrix, ...]:
        return self.symbolic_array_expressions


class LambdifiedVectorExpressions(LambdifiedMatrixExpressions):
    def __init__(
         self,
         symbolic_vector_expressions: tp.Tuple[sym.MatrixBase, ...],
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: tp.Optional[sym.Symbol] = None,
         symbolic_time_function_name_to_numeric_time_function_map:
         tp.Optional[tp.Dict[str, tp.Callable]] = None,
         squeeze_column_vectors: tp.Optional[tp.Tuple[bool, ...]] = None,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: tp.Optional[tp.Tuple[bool, ...]] = None,
         dtype: np.generic = np.float32):

        super().__init__(
                    symbolic_matrix_expressions=symbolic_vector_expressions,
                    argument_symbols=argument_symbols,
                    time_symbol=time_symbol,
                    symbolic_time_function_name_to_numeric_time_function_map
                    =symbolic_time_function_name_to_numeric_time_function_map,
                    squeeze_column_vectors=squeeze_column_vectors,
                    perform_cse=perform_cse,
                    lazy_initialization=lazy_initialization,
                    pre_attach=pre_attach, dtype=dtype)

    @property
    def dimension(self) -> int:
        return len(self.symbolic_vector_expressions)

    @property
    def symbolic_vector_expressions(self) -> tp.Tuple[sym.ImmutableMatrix, ]:
        return self.symbolic_matrix_expressions


################################################################################
# Single Lambdified Array, Matrix, and Vector-Expressions
################################################################################
class LambdifiedArrayExpression:
    def __init__(
         self,
         symbolic_array_expression: tp.Union[sym.MatrixBase, sym.Array],
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: tp.Optional[sym.Symbol] = None,
         symbolic_time_function_name_to_numeric_time_function_map:
         tp.Optional[tp.Dict[str, tp.Callable]] = None,
         squeeze_column_vector: bool = False,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: bool = True,
         dtype: np.generic = np.float32):

        self._lambdified_array_expressions = \
                LambdifiedArrayExpressions(
                    argument_symbols=argument_symbols,
                    time_symbol=time_symbol,
                    symbolic_array_expressions=(symbolic_array_expression, ),
                    symbolic_time_function_name_to_numeric_time_function_map
                    =symbolic_time_function_name_to_numeric_time_function_map,
                    squeeze_column_vectors=(squeeze_column_vector, ),
                    perform_cse=perform_cse,
                    lazy_initialization=lazy_initialization,
                    pre_attach=(pre_attach, ),
                    dtype=dtype)

    @property
    def is_column_vector(self) -> bool:
        return self._lambdified_array_expressions.is_column_vector(0)

    @property
    def is_row_vector(self) -> bool:
        return self._lambdified_array_expressions.is_row_vector(0)

    @property
    def argument_symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._lambdified_array_expressions.argument_symbols

    @property
    def number_of_state_variables(self) -> int:
        return self._lambdified_array_expressions.number_of_state_variables

    @property
    def time_symbol(self) -> sym.Symbol:
        return self._lambdified_array_expressions.time_symbol

    @property
    def symbolic_array_expression(self) -> sym.ImmutableMatrix:
        return self._lambdified_array_expressions.symbolic_array_expressions[0]

    @property
    def symbolic_time_function_name_to_numeric_time_function_map(self) \
            -> tp.Dict[str, tp.Callable]:
        return (self
                ._lambdified_array_expressions
                .symbolic_time_function_name_to_numeric_time_function_map)

    @property
    def symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._lambdified_array_expressions.symbols

    @property
    def number_of_variables(self) -> int:
        return self._lambdified_array_expressions.number_of_variables

    def evaluate_with_list_of_state_vectors(
            self,
            list_of_state_vectors: tp.Tuple[NumericArrayOrScalar, ...],
            t: tp.Optional[float] = None,
            gpu: tp.Optional[bool] = None) -> NumericArray:
        """
        This function evaluates the array expression at arguments that are
        vectors (arrays with a single axis) or scalars and returns the resulting
        array.

        Args:
            list_of_state_vectors:
                A list of one-dimensional numeric arrays. The length of this
                list must match the number of arguments to the array-valued
                functions.

            t: time parameter
            gpu: whether to evaluate the array expression on the GPU

        Returns: an array corresponding to the array expressions evaluated at
                 the parameter vectors

        """

        return (self
                ._lambdified_array_expressions
                .evaluate_with_list_of_state_vectors(list_of_state_vectors,
                                                     t,
                                                     gpu)[0])

    def evaluate_with_dictionary(
            self,
            symbolic_to_numeric_parameter_map: tp.Dict[sym.Symbol,
                                                       NumericArrayOrScalar],
            t: tp.Optional[float] = None,
            gpu: tp.Optional[bool] = None) -> NumericArray:
        """
        This function evaluates the array expression at arguments that are
        dictionaries mapping symbolic parameters to one-dimensional numeric
        arrays or scalars and returns the resulting array.

        Args:
            symbolic_to_numeric_parameter_map:
                A dictionary mapping symbolic parameters to numeric arrays or
                scalars.

            t: time parameter
            gpu: whether to evaluate the array expression on the GPU

        Returns: an array corresponding to the array expressions evaluated at
                 the parameter vectors

        """
        return (self
                ._lambdified_array_expressions
                .evaluate_with_dictionary(
                    symbolic_to_numeric_parameter_map
                    =symbolic_to_numeric_parameter_map,
                    t=t,
                    gpu=gpu))[0]

    def evaluate_with_kwargs(self,
                             t: tp.Optional[float] = None,
                             gpu: tp.Optional[bool] = None,
                             **kwargs) \
            -> NumericArray:
        """
        This function evaluates the array expression using the symbol names
        given by keyword arguments.

        Args:
            t: time parameter
            gpu: whether to evaluate the array expression on the GPU
            kwargs: a dictionary mapping parameter names to numeric arrays or
                    scalars

        Returns: an array corresponding to the array expressions evaluated at
                 the parameter vectors

        """
        return (self
                ._lambdified_array_expressions
                .evaluate_with_kwargs(t=t,
                                      gpu=gpu,
                                      **kwargs)[0])

    def evaluate(self,
                 state_matrices: tp.Tuple[NumericArrayOrScalar, ...],
                 t: tp.Optional[float] = None,
                 gpu: tp.Optional[bool] = None) -> NumericArray:
        """
        This function evaluates the array expression at arguments that are
        represented either as vectors or as horizontally concatenated vectors
        (i.e. arrays with 2 axes). It decomposes these matrices into column
        vectors and evaluates the array functions with this list of column vectors.

        Args:
            state_matrices: a tuple of matrices, vectors, or scalars
            t: time parameter
            gpu: whether to evaluate the array expression on the GPU

        Returns: an array corresponding to the array expressions
                 evaluated at the parameter vectors

        """
        return (self
                ._lambdified_array_expressions
                .evaluate(state_matrices=state_matrices,
                          t=t,
                          gpu=gpu)[0])


class LambdifiedMatrixExpression(LambdifiedArrayExpression):
    def __init__(
         self,
         symbolic_matrix_expression: sym.MatrixBase,
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: tp.Optional[sym.Symbol] = None,
         symbolic_time_function_name_to_numeric_time_function_map:
         tp.Optional[tp.Dict[str, tp.Callable]] = None,
         squeeze_column_vector: bool = False,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: bool = True,
         dtype: np.generic = np.float32):

        super().__init__(argument_symbols=argument_symbols,
                         time_symbol=time_symbol,
                         symbolic_array_expression=symbolic_matrix_expression,
                         symbolic_time_function_name_to_numeric_time_function_map
                         =symbolic_time_function_name_to_numeric_time_function_map,
                         squeeze_column_vector=squeeze_column_vector,
                         perform_cse=perform_cse,
                         lazy_initialization=lazy_initialization,
                         pre_attach=pre_attach,
                         dtype=dtype)

    @property
    def rows(self) -> int:
        return (self
                ._lambdified_array_expressions
                .symbolic_array_expressions[0]
                .shape[0])

    @property
    def cols(self) -> int:
        return (self
                ._lambdified_array_expressions
                .symbolic_array_expressions[0].shape[1])

    @property
    def symbolic_matrix_expression(self) -> sym.ImmutableMatrix:
        return self.symbolic_array_expression


class LambdifiedVectorExpression(LambdifiedMatrixExpression):
    def __init__(
         self,
         symbolic_vector_expression: sym.MatrixBase,
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: tp.Optional[sym.Symbol] = None,
         symbolic_time_function_name_to_numeric_time_function_map:
         tp.Optional[tp.Dict[str, tp.Callable]] = None,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: bool = True,
         dtype: np.generic = np.float32):

        super().__init__(argument_symbols=argument_symbols,
                         time_symbol=time_symbol,
                         symbolic_matrix_expression=symbolic_vector_expression,
                         symbolic_time_function_name_to_numeric_time_function_map
                         =symbolic_time_function_name_to_numeric_time_function_map,
                         squeeze_column_vector=True,
                         perform_cse=perform_cse,
                         lazy_initialization=lazy_initialization,
                         pre_attach=pre_attach,
                         dtype=dtype)

    @property
    def symbolic_vector_expression(self) -> sym.ImmutableMatrix:
        return self.symbolic_matrix_expression


class LambdifiedScalarExpression(LambdifiedVectorExpression):
    def __init__(
         self,
         symbolic_expression: sym.Expr,
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: tp.Optional[sym.Symbol] = None,
         symbolic_time_function_name_to_numeric_time_function_map:
         tp.Optional[tp.Dict[str, tp.Callable]] = None,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: bool = True,
         dtype: np.generic = np.float32):

        super().__init__(argument_symbols=argument_symbols,
                         time_symbol=time_symbol,
                         symbolic_vector_expression
                         =sym.ImmutableMatrix([symbolic_expression]),
                         symbolic_time_function_name_to_numeric_time_function_map
                         =symbolic_time_function_name_to_numeric_time_function_map,
                         perform_cse=perform_cse,
                         lazy_initialization=lazy_initialization,
                         pre_attach=pre_attach,
                         dtype=dtype)

    @property
    def symbolic_expression(self) -> sym.Expr:
        return self.symbolic_matrix_expression[0, 0]

    def evaluate_with_list_of_state_vectors(
            self,
            list_of_state_vectors: tp.Tuple[NumericArrayOrScalar, ...],
            t: tp.Optional[float] = None,
            gpu: tp.Optional[bool] = None) -> NumericArray:
        """
        This function evaluates the scalar expression at arguments that are
        vectors (arrays with a single axis) or scalars and returns the resulting
        array.

        Args:
            list_of_state_vectors:
                A list of one-dimensional numeric arrays. The length of this
                list must match the number of arguments to the array-valued
                functions.

            t: time parameter
            gpu: whether to evaluate the array expression on the GPU

        Returns: an array corresponding to the array expressions evaluated at
                 the parameter vectors

        """

        return (super()
                .evaluate_with_list_of_state_vectors(
                    list_of_state_vectors=list_of_state_vectors,
                    t=t,
                    gpu=gpu)
                .squeeze())

    def evaluate_with_dictionary(
            self,
            symbolic_to_numeric_parameter_map: tp.Dict[sym.Symbol,
                                                       NumericArrayOrScalar],
            t: tp.Optional[float] = None,
            gpu: tp.Optional[bool] = None) -> NumericArray:
        """
        This function evaluates the scalar expression at arguments that are
        dictionaries mapping symbolic parameters to one-dimensional numeric
        arrays or scalars and returns the resulting array.

        Args:
            symbolic_to_numeric_parameter_map:
                A dictionary mapping symbolic parameters to numeric arrays or
                scalars.

            t: time parameter
            gpu: whether to evaluate the array expression on the GPU

        Returns: an array corresponding to the array expressions evaluated at
                 the parameter vectors

        """
        return (super()
                .evaluate_with_dictionary(symbolic_to_numeric_parameter_map
                                          =symbolic_to_numeric_parameter_map,
                                          t=t,
                                          gpu=gpu)
                .squeeze())

    def evaluate_with_kwargs(self,
                             t: tp.Optional[float] = None,
                             gpu: tp.Optional[bool] = None,
                             **kwargs) \
            -> NumericArray:
        """
        This function evaluates the scalar expression using the symbol names
        given by keyword arguments.

        Args:
            t: time parameter
            gpu: whether to evaluate the array expression on the GPU
            kwargs: a dictionary mapping parameter names to numeric arrays or
                    scalars

        Returns: an array corresponding to the array expressions evaluated at
                 the parameter vectors

        """
        return super().evaluate_with_kwargs(t=t, gpu=gpu, **kwargs).squeeze()

    def evaluate(self,
                 state_matrices: tp.Tuple[NumericArrayOrScalar, ...],
                 t: tp.Optional[float] = None,
                 gpu: tp.Optional[bool] = None) -> NumericArray:
        """
        This function evaluates the scalar expression at arguments that are
        represented either as vectors or as horizontally concatenated vectors
        (i.e. arrays with 2 axes). It decomposes these matrices into column
        vectors and evaluates the array functions with this list of column
        vectors.

        Args:
            state_matrices: a tuple of matrices, vectors, or scalars
            t: time parameter
            gpu: whether to evaluate the array expression on the GPU

        Returns: an array corresponding to the array expressions
                 evaluated at the parameter vectors

        """
        return (super()
                .evaluate(state_matrices=state_matrices,
                          t=t,
                          gpu=gpu)
                .squeeze())


def lambdify(args,
             expr,
             modules: tp.Optional[tp.Tuple] = None,
             printer: tp.Optional[tp.Tuple] = None,
             use_imps: bool = True,
             dummify: bool = True):
    if modules is None:
        modules = []

    return sym.lambdify(args=args,
                        expr=expr,
                        modules=list(modules) + [COCOS_TRANSLATIONS],
                        printer=printer,
                        use_imps=use_imps,
                        dummify=dummify)



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.2, 0.25]],
                      dtype=np.float32),
             np.array([[0.5, 2.3, 3.1],
                       [4, 5.5, 6],
                       [7 - 9j, 8 + 1j, 2 + 10j]],
                      dtype=np.complex64),
             np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]], dtype=np.int32),
             np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32)
            ]


@pytest.mark.parametrize("A_numpy", test_data)
def test_squeeze_reshape(A_numpy):
    cocos.device.init()
    newshapes = [(3, 1, 3), (1, 3, 3)]
    axess = [(None, 1), (None, 0)]

    for newshape, axes in zip(newshapes, axess):
        A_cocos = cn.array(A_numpy)

        # 3, 1, 3
        B_numpy = A_numpy.reshape(newshape)
        B_cocos = A_cocos.reshape(newshape)

        assert np.allclose(B_cocos, B_numpy)

        for axis in axes:
            C_numpy = B_numpy.squeeze(axis=axis)
            C_cocos = B_cocos.squeeze(axis=axis)

            assert np.allclose(C_cocos, C_numpy)


def main():
    test_squeeze_reshape(test_data[0])


if __name__ == '__main__':
    main()



import collections
import numpy as np
from types import ModuleType
import typing as tp

from cocos.numerics.data_types import NumericArray
import cocos.numerics as cn


def select_num_pack(gpu: bool = True) -> ModuleType:
    if gpu:
        return cn
    else:
        return np


def select_num_pack_by_dtype(array: NumericArray) -> ModuleType:
    gpu, num_pack = get_gpu_and_num_pack_by_dtype(array)
    return num_pack


def get_gpu_and_num_pack_by_dtype(array: NumericArray) \
        -> tp.Tuple[bool, ModuleType]:
    if isinstance(array, cn.ndarray):
        return True, cn
    else:
        return False, np


def select_num_pack_by_dtype_from_iterable(arrays: tp.Iterable[NumericArray]) \
        -> ModuleType:
    gpu, num_pack = get_gpu_and_num_pack_by_dtype_from_iterable(arrays)
    return num_pack


def get_gpu_and_num_pack_by_dtype_from_iterable(
        arrays: tp.Iterable[NumericArray]) \
        -> tp.Tuple[bool, ModuleType]:
    if any([isinstance(array, cn.ndarray) for array in arrays]):
        return True, cn
    else:
        return False, np



import numpy as np
import pytest
import scipy as sp
import typing as tp

from cocos.scientific.kde import (
    gaussian_kernel_estimate,
    gaussian_kernel_estimate_vectorized
)

n = 1000


@pytest.fixture()
def points() -> np.ndarray:
    return np.random.randn(n, 1)


@pytest.fixture()
def weights(points: np.ndarray) -> np.ndarray:
    return np.full_like(points, fill_value=1.0 / n)


@pytest.fixture()
def xi() -> np.ndarray:
    return np.linspace(-5, 5, 101).reshape((-1, 1))


@pytest.fixture()
def precision(points: np.ndarray) -> np.ndarray:
    return np.atleast_2d(1.0 / np.var(points))


def test_gaussian_kernel_estimate_loop_vs_vectorized_cpu(points: np.ndarray,
                                                         weights: np.ndarray,
                                                         xi: np.ndarray,
                                                         precision: np.ndarray):
    kde = gaussian_kernel_estimate(points, weights, xi, precision, np.float64)
    kde_vectorized = np.array(gaussian_kernel_estimate_vectorized(points=points,
                                                                  values=weights,
                                                                  xi=xi,
                                                                  precision=precision,
                                                                  dtype=np.float64,
                                                                  gpu=False))

    np.allclose(kde_vectorized, kde)


def test_gaussian_kernel_estimate_loop_vs_vectorized_gpu(points: np.ndarray,
                                                         weights: np.ndarray,
                                                         xi: np.ndarray,
                                                         precision: np.ndarray):
    kde = gaussian_kernel_estimate(points, weights, xi, precision, np.float64)
    kde_vectorized = np.array(gaussian_kernel_estimate_vectorized(points=points,
                                                                  values=weights,
                                                                  xi=xi,
                                                                  precision=precision,
                                                                  dtype=np.float64,
                                                                  gpu=True))

    np.allclose(kde_vectorized, kde)



import numpy as np
from arrayfire import Dtype


def convert_numpy_to_af_type(numpy_type):
    if numpy_type == np.float32:
        return Dtype.f32
    elif numpy_type == np.complex64:
        return Dtype.c32
    elif numpy_type == np.float64:
        return Dtype.f64
    elif numpy_type == np.complex128:
        return Dtype.c64
    elif numpy_type == np.bool8 or \
            numpy_type == np.bool or \
            numpy_type == np.bool_:
        return Dtype.b8
    elif numpy_type == np.int32:
        return Dtype.s32
    elif numpy_type == np.uint32:
        return Dtype.u32
    elif numpy_type == np.uint8:
        return Dtype.u8
    elif numpy_type == np.int64:
        return Dtype.s64
    elif numpy_type == np.uint64:
        return Dtype.u64
    elif numpy_type == np.int or numpy_type == int:
        return Dtype.s32
    elif numpy_type == np.long:
        return Dtype.s64
    elif numpy_type == np.float or numpy_type == float:
        return Dtype.f32
    elif numpy_type == np.complex or numpy_type == complex:
        return Dtype.c32
    elif numpy_type == np.double:
        return Dtype.f64
    else:
        raise ValueError(f"Type {numpy_type} not supported.")


def convert_af_to_numpy_type(af_type):
    if af_type == Dtype.f32:
        return np.float32
    elif af_type == Dtype.c32:
        return np.complex64
    elif af_type == Dtype.f64:
        return np.float64
    elif af_type == Dtype.c64:
        return np.complex128
    elif af_type == Dtype.b8:
        return np.bool8
    # elif (af_type == Dtype.b8):
    #     return bool
    elif af_type == Dtype.s32:
        return np.int32
    elif af_type == Dtype.u32:
        return np.uint32
    elif af_type == Dtype.u8:
        return np.uint8
    elif af_type == Dtype.s64:
        return np.int64
    elif af_type == Dtype.u64:
        return np.uint64
    else:
        raise ValueError(f"Type {af_type} not supported.")



import math
import numpy
import time
import typing as tp

import cocos.device as cd
from cocos.numerics.numerical_package_selector import \
    get_gpu_and_num_pack_by_dtype

from cocos.numerics.random import randn_antithetic


def simulate_heston_model(T: float,
                          N: int,
                          R: int,
                          mu: float,
                          kappa: float,
                          v_bar: float,
                          sigma_v: float,
                          rho: float,
                          x0: float,
                          v0: float,
                          gpu: bool = False) -> tp.Tuple:
    """
    This function simulates R paths from the Heston stochastic volatility model
    over a time horizon of length T divided into N steps.

    :param T: time horizon of the simulation
    :param N: number of steps
    :param R: number of paths to simulate
    :param mu: expected return
    :param kappa: mean-reversion speed of volatility
    :param v_bar: long-run mean of volatility
    :param sigma_v: volatility of volatility
    :param rho: instantaneous correlation of shocks to price and to volatility
    :param x0: initial log price
    :param v0: initial volatility
    :param gpu: whether to compute on the GPU
    :return: a tuple of two R-dimensional numeric arrays for log price and
             volatility
    """
    if gpu:
        import cocos.numerics as np
    else:
        import numpy as np

    Delta_t = T / float(N - 1)

    x = [np.full((R,), x0, dtype=numpy.float32),
         np.zeros((R,), dtype=numpy.float32)]

    v = [np.full((R,), v0, dtype=numpy.float32),
         np.zeros((R,), dtype=numpy.float32)]

    sqrt_delta_t = math.sqrt(Delta_t)
    sqrt_one_minus_rho_square = math.sqrt(1 - rho ** 2)

    # m = np.array([[rho, sqrt_one_minus_rho_square]])
    m = np.zeros((2,), dtype=numpy.float32)
    m[0] = rho
    m[1] = sqrt_one_minus_rho_square
    zero_array = np.zeros((R,), dtype=numpy.float32)

    t_current = 0
    for t in range(1, N):
        t_previous = (t + 1) % 2
        t_current = t % 2

        # generate antithetic standard normal random variables
        dBt = randn_antithetic(shape=(R, 2),
                               antithetic_dimension=0,
                               num_pack=np,
                               dtype=numpy.float32) * sqrt_delta_t

        sqrt_v_lag = np.sqrt(v[t_previous])
        x[t_current] = x[t_previous] \
                     + (mu - 0.5 * v[t_previous]) * Delta_t \
                     + np.multiply(sqrt_v_lag, dBt[:, 0])
        v[t_current] = v[t_previous] \
                     + kappa * (v_bar - v[t_previous]) * Delta_t \
                     + sigma_v * np.multiply(sqrt_v_lag, np.dot(dBt, m))
        v[t_current] = np.maximum(v[t_current], numpy.finfo(numpy.float32).eps)

    x = x[t_current]
    v = np.maximum(v[t_current], numpy.finfo(numpy.float32).eps)

    return x, v


def compute_option_price(r: float,
                         T: float,
                         K: float,
                         x_simulated,
                         num_pack: tp.Optional = None):
    """
    Compute the function of a plain-vanilla call option from simulated
    log-returns.

    :param r: the risk-free rate
    :param T: the time to expiration
    :param K: the strike price
    :param x_simulated: a numeric array of simulated log prices of the underlying
    :param num_pack: a module - either numpy or cocos.numerics
    :return: option price
    """

    if num_pack is None:
        use_gpu, num_pack = get_gpu_and_num_pack_by_dtype(x_simulated)

    return math.exp(-r * T) \
           * num_pack.mean(num_pack.maximum(num_pack.exp(x_simulated) - K, 0))


def run_benchmark(x0: float,
                  v0: float,
                  r: float,
                  rho: float,
                  sigma_v: float,
                  kappa: float,
                  v_bar: float,
                  T: float,
                  K: float,
                  nT: int,
                  R: int,
                  use_gpu: bool = True) -> tp.Tuple[float, float, float, float]:
    if use_gpu:
        import cocos.numerics as np
    else:
        import numpy as np

    # number of paths for warm-up (compile GPU kernels)
    R_warm_up = 1000
    x_simulated, v_simulated \
        = simulate_heston_model(T=T,
                                N=nT,
                                R=R_warm_up,
                                mu=r,
                                kappa=kappa,
                                v_bar=v_bar,
                                sigma_v=sigma_v,
                                rho=rho,
                                x0=x0,
                                v0=v0,
                                gpu=use_gpu)

    # actual simulation run to price plain vanilla call option
    tic = time.time()
    (x_simulated, v_simulated) \
        = simulate_heston_model(T=T,
                                N=nT,
                                R=R,
                                mu=r,
                                kappa=kappa,
                                v_bar=v_bar,
                                sigma_v=sigma_v,
                                rho=rho,
                                x0=x0,
                                v0=v0,
                                gpu=use_gpu)

    if use_gpu:
        cd.sync()
    time_in_simulation = time.time() - tic

    tic = time.time()

    # compute option price
    option_price = compute_option_price(r, T, K, x_simulated, num_pack=None)

    time_in_option_price_calculation = time.time() - tic

    total_time = time_in_simulation + time_in_option_price_calculation

    return (time_in_simulation,
            time_in_option_price_calculation,
            total_time,
            option_price)


def main():
    cd.info()

    # model parameters
    x0 = 0.0  # initial log stock price
    v0 = 0.101 ** 2  # initial volatility
    r = math.log(1.0319)  # risk-free rate
    rho = -0.7  # instantaneous correlation between Brownian motions
    sigma_v = 0.61  # variance of volatility
    kappa = 6.21  # mean reversion speed
    v_bar = 0.019  # mean variance

    # option parameters
    T = 1.0  # time to expiration
    K = 0.95  # strike price

    # simulation parameters
    nT = int(math.ceil(500 * T))  # number of time-steps to simulate
    R = 2000000  # actual number of paths to simulate for pricing

    # run benchmark on cpu
    (time_in_simulation_cpu,
     time_in_option_price_calculation_cpu,
     total_time_cpu,
     option_price_cpu) \
        = run_benchmark(x0,
                        v0,
                        r,
                        rho,
                        sigma_v,
                        kappa,
                        v_bar,
                        T,
                        K,
                        nT,
                        R,
                        use_gpu=False)

    print(f"Time in simulation on CPU: {time_in_simulation_cpu} secs")
    print(f"Time in option price calculation on CPU: "
          f"= {time_in_option_price_calculation_cpu} secs")

    print(f"Total time on CPU: "
          f"= {total_time_cpu} secs")
    print(f"Call option price on CPU: {option_price_cpu}")

    # run benchmark on gpu
    (time_in_simulation_gpu,
     time_in_option_price_calculation_gpu,
     total_time_gpu,
     option_price_gpu) \
        = run_benchmark(x0,
                        v0,
                        r,
                        rho,
                        sigma_v,
                        kappa,
                        v_bar,
                        T,
                        K,
                        nT,
                        R,
                        use_gpu=True)

    print(f"Time in simulation on GPU: {time_in_simulation_gpu} secs")
    print(f"Time in option price calculation on GPU: "
          f"= {time_in_option_price_calculation_gpu} secs")

    print(f"Total time on GPU: "
          f"= {total_time_gpu} secs")
    print(f"Call option price on GPU: {option_price_gpu}")

    # compute and print the speedup factor
    print(f'Speedup: {total_time_cpu/ total_time_gpu}')


if __name__ == "__main__":
    main()



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [(np.array([[True, True, False],
                        [True, False, True],
                        [False, True, True]],
                       dtype=bool),
             np.array([[True, False, True],
                       [False, False, True],
                       [True, True, True]],
                      dtype=bool))]


@pytest.mark.parametrize("A_numpy, B_numpy", test_data)
def test_logical(A_numpy, B_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)
    B_cocos = cn.array(B_numpy)

    assert np.allclose(cn.logical_not(A_cocos),
                       np.logical_not(A_numpy))

    assert np.allclose(cn.logical_and(A_cocos, B_cocos),
                       np.logical_and(A_numpy, B_numpy))

    assert np.allclose(cn.logical_or(A_cocos, B_cocos),
                       np.logical_or(A_numpy, B_numpy))

    assert np.allclose(cn.logical_xor(A_cocos, B_cocos),
                       np.logical_xor(A_numpy, B_numpy))



import arrayfire as af
import typing as tp

from ._array import ndarray, _wrap_af_array


def count_nonzero(a: ndarray,
                  axis: tp.Optional[int] = None) \
        -> tp.Union[int, ndarray]:
    return _wrap_af_array(af.count(a._af_array, dim=axis))


def diff(a: ndarray,
         n: int = 1,
         axis: int = -1) -> ndarray:
    """Calculate the n-th discrete difference along given axis."""

    if axis == -1:
        # use last axis
        axis = a.ndim - 1

    if 0 <= axis <= 3:
        if axis >= a.ndim:
            raise ValueError("axis exceeds array dimension")

        if n >= a.shape[axis]:
            raise ValueError(f"input array has length {a.shape[axis]} in "
                             f"dimension {axis} and therefore cannot be "
                             f"differentiated more than {a.shape[axis] - 1} "
                             f"times")

        if n == 0:
            return a.copy()
        elif n == 1:
            new_array = af.diff1(a._af_array, dim=axis)
        elif n == 2:
            new_array = af.diff2(a._af_array, dim=axis)
        elif n > 2:
            output = a
            while n >= 2:
                n -= 2
                output = ndarray(af.diff2(output._af_array, dim=axis))

            if n == 1:
                output = ndarray(af.diff1(output._af_array, dim=axis))

            return output
        else:
            raise ValueError(f"n must be positive but is {n}")
    else:
        raise ValueError("Axis must be between 0 and 3")

    return ndarray(new_array)


def flatnonzero(a: ndarray) -> ndarray:
    return ndarray(af.where(a._af_array))


def sort_by_keys(keys: ndarray,
                 values: ndarray,
                 axis: int = -1,
                 ascending: bool = True) -> tp.Tuple[ndarray, ndarray]:
    if keys.shape != values.shape:
        raise ValueError("Keys and values must have the same dimensions.")
    elif axis is None:
        keys = keys.flatten()
        values = values.flatten()
    elif axis == -1:
        axis = keys.ndim - 1
    elif axis >= keys.ndim:
        raise ValueError(f"Parameter axis must be between -1 and "
                         f"{keys.ndim - 1}")

    ordered_values, ordered_keys \
        = af.sort_by_key(values._af_array,
                         keys._af_array,
                         is_ascending=ascending)

    return ndarray(ordered_keys), ndarray(ordered_values)


def unique(ar: ndarray,
           return_index: bool = False,
           return_inverse: bool = False,
           return_counts: bool = False) -> ndarray:
    if return_index:
        raise ValueError("return_index=True is not supported")
    if return_inverse:
        raise ValueError("return_inverse=True is not supported")
    if return_counts:
        raise ValueError("return_counts=True is not supported")

    unsorted_unique_set_af_array = af.set_unique(ar._af_array,
                                                 is_sorted=False)

    sorted_unique_set_af_array = af.sort(unsorted_unique_set_af_array,
                                         dim=0,
                                         is_ascending=True)

    return ndarray(sorted_unique_set_af_array)


def union1d(ar1: ndarray, ar2: ndarray) -> ndarray:
    new_af_array = af.set_union(ar1._af_array,
                                ar2._af_array,
                                is_unique=False)

    return ndarray(new_af_array)


def intersect1d(ar1: ndarray, ar2: ndarray) -> ndarray:
    new_af_array = af.set_intersect(ar1._af_array,
                                    ar2._af_array,
                                    is_unique=False)

    return ndarray(new_af_array)



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest

y_numpy = np.array([[1, 2, 3], [4, 5, 6]])
y_cocos = cn.array(y_numpy)


@pytest.mark.parametrize('axis', (0, 1))
def test_repeat(axis: int):
    pass
    # result_numpy = np.repeat(y_numpy, repeats=2, axis=axis)
    # result_cocos = cn.repeat(y_cocos, repeats=2, axis=axis)
    #
    # assert np.all(result_numpy == np.array(result_cocos))



from ._lambdification import (
    DUMMY_TIME_SYMBOL,
    lambdify,
    LambdifiedArrayExpressions,
    LambdifiedMatrixExpressions,
    LambdifiedVectorExpressions,
    LambdifiedArrayExpression,
    LambdifiedMatrixExpression,
    LambdifiedVectorExpression,
    LambdifiedScalarExpression
)

from .utilities import find_length_of_state_vectors



import typing as tp

from cocos.device import sync
from cocos.multi_processing.utilities import (
    MultiprocessingPoolType,
    ResultType,
    ParameterTransferFunction,
    _extract_arguments_and_number_of_batches
)


def map_reduce_single_gpu(
        f: tp.Callable[..., ResultType],
        reduction: tp.Callable[[ResultType, ResultType], ResultType],
        initial_value: ResultType,
        host_to_device_transfer_function:
        tp.Optional[ParameterTransferFunction] = None,
        device_to_host_transfer_function:
        tp.Optional[tp.Callable[[ResultType], ResultType]] = None,
        args_list: tp.Optional[tp.Sequence[tp.Sequence]] = None,
        kwargs_list: tp.Optional[tp.Sequence[tp.Dict[str, tp.Any]]] = None,
        number_of_batches: tp.Optional[int] = None) \
        -> ResultType:
    """
    This method function the function 'f' on elements of 'args_list' and 
    'kwargs_list' sequentially on a single device and performs the reduction 
    by calling the function 'reduction' on the result and the result of the 
    reductions so far to eventually produce one final result of type 
    'ResultType'. The reduce step is performed from the left and results are 
    being processed in the same order as they appear in `args_list` and 
    `kwargs_list`. 

    Input data to the function f must initially reside in host memory and 
    the user must provide functions 'host_to_device_transfer_function' and 
    'device_to_host_transfer_function' to transfer the data to and results 
    from device memory respectively.

    If the arguments for each run of 'f' are identical and they have already 
    been applied to the function that is passed then 'args_list' and 
    'kwargs_list' may both be None but the argument 'number_of_batches' must 
    be specified so the method knows how many times to run the function 'f'.
    
    Args:
        f: The map function to be evaluated over elements of 'args_list' and 
           'kwargs_list'.
           
        reduction: The reduction to be performed on the results of 'f'. 
                   This is done on the host (not the device).
                   
        initial_value: The initial value of the reduction 
                       (i.e. the neutral element).
                       
        host_to_device_transfer_function: 
            A function that transfers elements of args_list and kwargs_list 
            from host memory to device memory.
            
        device_to_host_transfer_function: 
            A function that transfers results from device to host memory.
            
        args_list: A sequence of sequences of positional arguments.
        kwargs_list: A sequence of dictionaries of keyword arguments.
        number_of_batches: 
            The number of function evaluations is required if 'args_list' 
            and 'kwargs_list' are both empty.
    """
    args_list, kwargs_list, number_of_batches = \
        _extract_arguments_and_number_of_batches(
            args_list=args_list,
            kwargs_list=kwargs_list,
            number_of_batches=number_of_batches)

    result = initial_value

    for args, kwargs in zip(args_list, kwargs_list):
        if host_to_device_transfer_function is not None:
            args, kwargs = host_to_device_transfer_function(*args, **kwargs)
        sync()
        new_part = f(*args, **kwargs)
        if device_to_host_transfer_function is not None:
            new_part = device_to_host_transfer_function(new_part)
        sync()

        result = reduction(result, new_part)

    return result


def map_combine_single_gpu(
        f: tp.Callable[..., ResultType],
        combination: tp.Callable[[tp.Iterable[ResultType]], ResultType],
        host_to_device_transfer_function:
        tp.Optional[ParameterTransferFunction] = None,
        device_to_host_transfer_function:
        tp.Optional[tp.Callable[[ResultType], ResultType]] = None,
        args_list: tp.Optional[tp.Sequence[tp.Sequence]] = None,
        kwargs_list: tp.Optional[tp.Sequence[tp.Dict[str, tp.Any]]] = None,
        number_of_batches: tp.Optional[int] = None) \
        -> ResultType:
    """
    This function evaluates the function `f` on elements of `args_list` and 
    `kwargs_list` sequentially on a single device and aggregates results 
    in a single step by calling the function `combination` with a list of all 
    results. Results provided to `combination` are in the same order as 
    they appear in `args_list` and `kwargs_list`. 
    
    Input data to the function f must initially reside in host memory and 
    the user must provide functions 'host_to_device_transfer_function' and 
    'device_to_host_transfer_function' to transfer the data to and results 
    from device memory respectively.
    
    If the arguments for each run of 'f' are identical and they have already 
    been applied to the function that is passed then 'args_list' and 
    'kwargs_list' may both be None but the argument 'number_of_batches' must 
    be specified so the method knows how many times to run the function 'f'.
    
    Args:
        f: The map function to be evaluated over elements of 'args_list' and 
           'kwargs_list'.
           
        combination: 
            A function that aggregates a list of all results in a single step
            
        host_to_device_transfer_function: 
            A function that transfers elements of args_list and kwargs_list 
            from host memory to device memory.
            
        device_to_host_transfer_function:
             A function that transfers results from device to host memory.
             
        args_list: A sequence of sequences of positional arguments.
        kwargs_list: A sequence of dictionaries of keyword arguments.
        number_of_batches: 
            The number of function evaluations is required if 'args_list' 
            and 'kwargs_list' are both empty.
    """
    args_list, kwargs_list, number_of_batches = \
        _extract_arguments_and_number_of_batches(
            args_list=args_list,
            kwargs_list=kwargs_list,
            number_of_batches=number_of_batches)

    results = []

    for args, kwargs in zip(args_list, kwargs_list):
        if host_to_device_transfer_function is not None:
            args, kwargs = host_to_device_transfer_function(*args, **kwargs)
        sync()
        result = f(*args, **kwargs)
        if device_to_host_transfer_function is not None:
            result = device_to_host_transfer_function(result)
        sync()
        results.append(result)

    return combination(results)



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.2, 0.25]],
                      dtype=np.float32),
             np.array([[0.5, 2.3, 3.1],
                       [4, 5.5, 6],
                       [7 - 9j, 8 + 1j, 2 + 10j]],
                      dtype=np.complex64),
             np.array([[[1.0, 2], [3, 4]],
                       [[5, 6], [7, 8]]],
                      dtype=np.float32)]

test_axes = (0, 1)
test_offset = (-2, -1, 0, 1, 2)


@pytest.mark.parametrize("A_numpy", test_data)
@pytest.mark.parametrize("axis", test_axes)
@pytest.mark.parametrize("shift", test_offset)
def test_roll(A_numpy, axis, shift):
    print(f'axis={axis}, shift={shift}')
    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    B_numpy = np.roll(A_numpy, shift=shift, axis=axis)
    B_cocos = cn.roll(A_cocos, shift=shift, axis=axis)
    assert np.allclose(B_cocos, B_numpy)



import numpy as np
import pytest
import scipy.stats as ss

import cocos.numerics as cn
from cocos.scientific.kde import gaussian_kde

n = 1000
grid_size = 101


points_1d = np.random.randn(n)
xi_1d = np.linspace(-5, 5, grid_size)


def generate_points_2d():
    points_x = np.random.randn(1, n)
    points_y = np.random.randn(1, n)
    return np.vstack((points_x, points_y))


def generate_xi_2d():
    x_grid = np.linspace(-5.0, 5.0, grid_size)
    y_grid = np.linspace(-5.0, 5.0, grid_size)
    xy_grid_x, xy_grid_y = np.meshgrid(x_grid, y_grid)
    return np.hstack((xy_grid_x.reshape((-1, 1)), xy_grid_y.reshape((-1, 1)))).T


points_2d = generate_points_2d()
xi_2d = generate_xi_2d()


@pytest.mark.parametrize('points, xi', [(points_1d, xi_1d),
                                        (points_2d, xi_2d)])
def test_gaussian_kde_scipy_vs_cocos(points: np.ndarray,
                                     xi: np.ndarray):
    gkde_cocos = gaussian_kde(points)
    gkde_scipy = ss.kde.gaussian_kde(points)

    density_estimate_cocos = gkde_cocos.evaluate(xi)
    density_estimate_scipy = gkde_scipy.evaluate(xi)

    assert np.allclose(density_estimate_cocos,
                       density_estimate_scipy)


@pytest.mark.parametrize('points, xi', [(points_1d, xi_1d),
                                        (points_2d, xi_2d)])
def test_gaussian_kde_scipy_vs_cocos_gpu(points: np.ndarray,
                                         xi: np.ndarray):
    gkde_cocos = gaussian_kde(cn.array(points.squeeze()),
                              gpu=True)
    gkde_scipy = ss.kde.gaussian_kde(points)

    density_estimate_cocos = gkde_cocos.evaluate(xi)
    density_estimate_scipy = gkde_scipy.evaluate(xi)

    assert np.allclose(density_estimate_cocos,
                       density_estimate_scipy)



import numpy as np

from cocos.options import GPUOptions
import cocos.numerics.random as random
import cocos.numerics.linalg as linalg

if GPUOptions.use_gpu:
    # from arrayfire/algorithm.py
    from ._algorithm import \
        count_nonzero, \
        diff, \
        flatnonzero, \
        intersect1d, \
        sort_by_keys, \
        union1d, \
        unique

    from ._array import \
        add, \
        all, \
        any, \
        argmax, \
        argmin, \
        argsort, \
        divide, \
        cumsum, \
        floor_divide, \
        multiply, \
        prod, \
        sort, \
        sort_argsort, \
        subtract, \
        sum, \
        true_divide

    # from arrayfire/_array.py
    from ._array import \
        ndarray, \
        array, \
        transpose, \
        display, \
        reorder, \
        reshape, \
        reshape_without_reorder, \
        squeeze, \
        nonzero, \
        trace, \
        reciprocal
    from cocos.device import init, info, sync, is_dbl_supported, selected_backend

    from ._array import \
        isempty, \
        isscalar, \
        isrow, \
        iscolumn, \
        isvector, \
        isrealobj, \
        iscomplexobj, \
        isdouble, \
        issingle, \
        isrealfloating, \
        isfloating, \
        isinteger, \
        isbool
    from ._array import \
        array_equal, \
        greater, \
        greater_equal, \
        less, \
        less_equal, \
        equal, \
        not_equal

    # from arrayfire/_arith.py
    from ._arith import \
        abs, \
        absolute, \
        minimum, \
        maximum, \
        sqrt, \
        square, \
        exp, \
        log, \
        iszero, \
        isinf, \
        isnan, \
        logical_and, \
        logical_or, \
        logical_not, \
        logical_xor, \
        bitwise_and, \
        bitwise_or, \
        invert, \
        bitwise_xor, \
        left_shift, \
        right_shift, \
        remainder, \
        mod, \
        angle, \
        sign, \
        trunc, \
        ceil, \
        hypot, \
        sin, \
        cos, \
        tan, \
        arcsin, \
        arccos, \
        arctan, \
        arctan2, \
        cplx, \
        sinh, \
        cosh, \
        tanh, \
        arcsinh, \
        arccosh, \
        arctanh, \
        root, \
        power, \
        power2, \
        expit, \
        expm1, \
        erf, \
        erfc, \
        log1p, \
        log1p, \
        log10, \
        log2, \
        cbrt, \
        factorial, \
        gamma, \
        gammaln, \
        isfinite, \
        isneginf, \
        isposinf

    from ._array import \
        conj, \
        floor, \
        real, \
        imag, \
        round, \
        clip

    # from arrayfire/blas.py
    from ._array import dot

    # from arrayfire/data.py
    from ._array import (
        full,
        tile,
        # repeat
    )

    from ._data import \
        empty, \
        zeros, \
        ones, \
        empty_like, \
        zeros_like, \
        ones_like, \
        eye, \
        identity, \
        diag, \
        diagflat, \
        concatenate, \
        vstack, \
        hstack, \
        dstack, \
        roll, \
        fliplr, \
        flipud, \
        tril, \
        triu

    # from arrayfire/statistics.py
    from ._array import \
        average, \
        corrcoef, \
        cov, \
        median, \
        mean, \
        std, \
        var

    from . import random

    init(None)
else:
    def is_dbl_supported() -> bool:
        return True

    class ndarray(np.ndarray):
        @property
        def label(self) -> str:
            return self._label

        @label.setter
        def label(self, label):
            self._label = label

    def init(backend: str=None):
        pass

    def sync(device=None):
        pass

    def info():
        print("dummy gpu numerics emulated on cpu")

    def selected_backend() -> str:
        return "dummy numpy cpu"

    def isempty(num) -> bool:
        raise NotImplementedError()

    def isrow(num) -> bool:
        raise NotImplementedError()

    def iscolumn(num) -> bool:
        raise NotImplementedError()

    def isdouble(num) -> bool:
        raise NotImplementedError()

    def issingle(num) -> bool:
        raise NotImplementedError()

    def isrealfloating(num) -> bool:
        raise NotImplementedError()

    def isfloating(num) -> bool:
        raise NotImplementedError()

    def isinteger(num) -> bool:
        raise NotImplementedError()

    def isbool(num) -> bool:
        raise NotImplementedError()

    def iszero(num) -> bool:
        raise NotImplementedError()

    def cplx(a, b=None):
        raise NotImplementedError()

    def root(a, b):
        raise NotImplementedError()

    def power2(a):
        raise NotImplementedError()

    # from arrayfire/algorithm.py
    from numpy import \
        all, \
        any, \
        argmax, \
        argmin, \
        argsort, \
        cumsum, \
        multiply, \
        prod, \
        sort, \
        sum

    # from arrayfire/_array.py
    from numpy import \
        ndarray, \
        array, \
        transpose, \
        reshape, \
        squeeze, \
        nonzero, \
        trace, \
        reciprocal
    from numpy import \
        array_equal, \
        greater, \
        greater_equal, \
        less, \
        less_equal, \
        equal, \
        not_equal

    # from arrayfire/_arith.py

    # from arrayfire/blas.py
    from numpy import dot

    # from arrayfire/data.py

    # from arrayfire/statistics.py

    # from numpy import array, transpose, reshape, nonzero, trace, reciprocal, roll
    #
    # from numpy import absolute, minimum, maximum, sqrt, square, exp, log, isinf, isnan, logical_and, logical_or, logical_not, logical_xor, \
    #     bitwise_and, \
    #     bitwise_or, invert, bitwise_xor, left_shift, right_shift, remainder, mod, angle, sign, trunc, floor, ceil, \
    #     hypot, sin, cos, tan, arcsin, arccos, arctan, arctan2, real, imag, conj, sinh, cosh, tanh, arcsinh, arccosh, arctanh, power, \
    #     expm1, log1p, log10, log2, cbrt, isfinite

    # from numpy import dot

# ARCHIMEDES = {}
#
# ARCHIMEDES_DEFAULT = {"I": 1j}
#
# ARCHIMEDES_TRANSLATIONS = {
#     "acos": "arccos",
#     "acosh": "arccosh",
#     "arg": "angle",
#     "asin": "arcsin",
#     "asinh": "arcsinh",
#     "atan": "arctan",
#     "atan2": "arctan2",
#     "atanh": "arctanh",
#     "ceiling": "ceil",
#     "E": "e",
#     "im": "imag",
#     "ln": "log",
#     "Mod": "mod",
#     "oo": "inf",
#     "re": "real",
#     "SparseMatrix": "array",
#     "ImmutableSparseMatrix": "array",
#     "Matrix": "array",
#     "MutableDenseMatrix": "array",
#     "ImmutableMatrix": "array",
#     "ImmutableDenseMatrix": "array",
# }
#
# ARCHIMEDES_SYMPY = (ARCHIMEDES, ARCHIMEDES_DEFAULT, ARCHIMEDES_TRANSLATIONS, ("import_module('archimedes.numerics')",))



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest

# backend = 'cpu'
# backend = 'cuda'
# backend = 'opencl'
backend = None

test_data = [np.array([[1, 2, 3], [4, 5, 6], [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5], [0.4, 0.5, 0.6], [0.7, 0.2, 0.25]],
                      dtype=np.float32)]


@pytest.mark.parametrize("A", test_data)
def test_var(A):
    cocos.device.init(backend)
    cocos.device.info()
    A_arch = cn.array(A)

    # tests variance
    assert np.allclose(np.var(A), cn.var(A_arch))
    assert np.allclose(np.var(A, axis=0), cn.var(A_arch, axis=0))
    assert np.allclose(np.var(A, axis=1), cn.var(A_arch, axis=1))

    # tests standard deviation
    assert np.allclose(np.std(A), cn.std(A_arch))
    assert np.allclose(np.std(A, axis=0), cn.std(A_arch, axis=0))
    assert np.allclose(np.std(A, axis=1), cn.std(A_arch, axis=1))

    # tests median
    assert np.allclose(np.median(A), cn.median(A_arch))
    assert np.allclose(np.median(A, axis=0), cn.median(A_arch, axis=0))
    assert np.allclose(np.median(A, axis=1), cn.median(A_arch, axis=1))

    # tests covariance
    a = np.cov(A)
    # b = cn.cov(A_arch)
    # assert np.allclose(np.cov(A), cn.cov(A_arch))
    # assert np.allclose(np.median(A, axis=0), cn.median(A_arch, axis=0))
    # assert np.allclose(np.median(A, axis=1), cn.median(A_arch, axis=1))



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest

test_data = [np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.2, 0.25]],
                      dtype=np.float32),
             np.array([[0.5, 2.3, 3.1],
                       [4, 5.5, 6],
                       [7 - 9j, 8 + 1j, 2 + 10j]],
                      dtype=np.complex64)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_flip(A_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    # conduct tests
    assert np.allclose(np.fliplr(A_numpy), cn.fliplr(A_cocos))
    assert np.allclose(np.flipud(A_numpy), cn.flipud(A_cocos))



import math
import matplotlib.pyplot as plt
import multiprocessing
import numpy
import time
import typing as tp

from cocos.device import (
    ComputeDeviceManager,
    info,
    sync
)

from cocos.multi_processing.device_pool import ComputeDevicePool
from cocos.multi_processing.multi_core_batch_processing import map_reduce_multicore
from cocos.numerics.numerical_package_bundle import (
    NumericalPackageBundle,
    CocosBundle
)

from cocos.numerics.random import randn_antithetic


def simulate_heston_model(
        T: float,
        N: int,
        R: int,
        mu: float,
        kappa: float,
        v_bar: float,
        sigma_v: float,
        rho: float,
        x0: float,
        v0: float,
        numerical_package_bundle: tp.Type[NumericalPackageBundle]) \
        -> tp.Tuple:
    """
    This function simulates R paths from the Heston stochastic volatility model
    over a time horizon of length T divided into N steps.

    :param T: time horizon of the simulation
    :param N: number of steps
    :param R: number of paths to simulate
    :param mu: expected return
    :param kappa: mean-reversion speed of volatility
    :param v_bar: long-run mean of volatility
    :param sigma_v: volatility of volatility
    :param rho: instantaneous correlation of shocks to price and to volatility
    :param x0: initial log price
    :param v0: initial volatility
    :param gpu: whether to compute on the GPU
    :return: a tuple of two R-dimensional numeric arrays for log price and
             volatility
    """
    np = numerical_package_bundle.module()
    random = numerical_package_bundle.random_module()

    Delta_t = T / float(N - 1)

    x = [np.full((R,), x0, dtype=numpy.float32),
         np.zeros((R,), dtype=numpy.float32)]

    v = [np.full((R,), v0, dtype=numpy.float32),
         np.zeros((R,), dtype=numpy.float32)]

    sqrt_delta_t = math.sqrt(Delta_t)
    sqrt_one_minus_rho_square = math.sqrt(1 - rho ** 2)

    m = np.zeros((2,), dtype=numpy.float32)
    m[0] = rho
    m[1] = sqrt_one_minus_rho_square

    t_current = 0
    for t in range(1, N):
        t_previous = (t + 1) % 2
        t_current = t % 2

        # generate antithetic standard normal random variables
        dBt = randn_antithetic(shape=(R, 2),
                               antithetic_dimension=0,
                               num_pack=np,
                               dtype=numpy.float32) * sqrt_delta_t

        sqrt_v_lag = np.sqrt(v[t_previous])
        x[t_current] = x[t_previous] \
                       + (mu - 0.5 * v[t_previous]) * Delta_t \
                       + np.multiply(sqrt_v_lag, dBt[:, 0])
        v[t_current] = v[t_previous] \
                       + kappa * (v_bar - v[t_previous]) * Delta_t \
                       + sigma_v * np.multiply(sqrt_v_lag, np.dot(dBt, m))
        v[t_current] = np.maximum(v[t_current], numpy.finfo(numpy.float32).eps)

    x = x[t_current]
    v = np.maximum(v[t_current], numpy.finfo(numpy.float32).eps)

    return x, v


def compute_option_price_from_simulated_paths(
        r: float,
        T: float,
        K: float,
        x_simulated,
        numerical_package_bundle: tp.Type[NumericalPackageBundle]) -> float:
    """
    Compute the function of a plain-vanilla call option from simulated
    log-returns.

    :param r: the risk-free rate
    :param T: the time to expiration
    :param K: the strike price
    :param x_simulated: a numeric array of simulated log prices of the underlying
    :param numerical_package_bundle: a class implementing NumericalPackageBundle
    :return: option price
    """

    num_pack = numerical_package_bundle.module()

    return float(math.exp(-r * T) * num_pack.mean(num_pack.maximum(num_pack.exp(x_simulated) - K, 0)))


def simulate_and_compute_option_price(
        x0: float,
        v0: float,
        r: float,
        rho: float,
        sigma_v: float,
        kappa: float,
        v_bar: float,
        T: float,
        K: float,
        nT: int,
        R: int,
        numerical_package_bundle: tp.Type[NumericalPackageBundle],
        verbose: bool = False) -> float:

    if verbose:
        print(f'computing on device={ComputeDeviceManager.get_current_compute_device_id()}')

    # simulate random paths
    x_simulated, v_simulated \
        = simulate_heston_model(
            T=T,
            N=nT,
            R=R,
            mu=r,
            kappa=kappa,
            v_bar=v_bar,
            sigma_v=sigma_v,
            rho=rho,
            x0=x0,
            v0=v0,
            numerical_package_bundle=numerical_package_bundle)

    # compute option price
    option_price \
        = compute_option_price_from_simulated_paths(
            r=r,
            T=T,
            K=K,
            x_simulated=x_simulated,
            numerical_package_bundle=numerical_package_bundle)

    return option_price


def simulate_and_compute_option_price_multicore(
        x0: float,
        v0: float,
        r: float,
        rho: float,
        sigma_v: float,
        kappa: float,
        v_bar: float,
        T: float,
        K: float,
        nT: int,
        R: int,
        numerical_package_bundle: tp.Type[NumericalPackageBundle],
        number_of_cores: tp.Optional[int] = None) -> float:
    if number_of_cores is None:
        # number_of_cores = 5 * multiprocessing.cpu_count()
        number_of_cores = 5 * multiprocessing.cpu_count()

    kwargs = \
        dict(x0=x0,
             v0=v0,
             r=r,
             rho=rho,
             sigma_v=sigma_v,
             kappa=kappa,
             v_bar=v_bar,
             T=T,
             K=K,
             nT=nT,
             numerical_package_bundle=numerical_package_bundle)

    # actual simulation run to price plain vanilla call option
    if number_of_cores == 1:
        kwargs['R'] = R
        option_price = \
            simulate_and_compute_option_price(
                **kwargs)

    else:
        kwargs['R'] = math.ceil(R / number_of_cores)

        option_price = \
            map_reduce_multicore(f=simulate_and_compute_option_price,
                                 reduction=lambda x, y: x + y / number_of_cores,
                                 initial_value=0.0,
                                 kwargs_list=number_of_cores * [kwargs])

    return option_price


def simulate_and_compute_option_price_gpu(
        x0: float,
        v0: float,
        r: float,
        rho: float,
        sigma_v: float,
        kappa: float,
        v_bar: float,
        T: float,
        K: float,
        nT: int,
        R: int,
        gpu_pool: tp.Optional[ComputeDevicePool] = None,
        number_of_batches: tp.Optional[int] = None) -> float:
    numerical_package_bundle = CocosBundle
    if number_of_batches is None:
        if gpu_pool is None:
            number_of_batches = 1
        else:
            number_of_batches = gpu_pool.number_of_devices

    kwargs = \
        dict(x0=x0,
             v0=v0,
             r=r,
             rho=rho,
             sigma_v=sigma_v,
             kappa=kappa,
             v_bar=v_bar,
             T=T,
             K=K,
             nT=nT,
             numerical_package_bundle=numerical_package_bundle)

    if gpu_pool is None:
        kwargs['R'] = R
        print(f'computing {kwargs["R"]} paths on single GPU')

        option_price = \
            simulate_and_compute_option_price(
                **kwargs)

    else:
        kwargs['R'] = math.ceil(R / number_of_batches)
        print(f'computing {R} paths on {gpu_pool.number_of_devices} GPUs in '
              f'{number_of_batches} batches of {kwargs["R"]} paths')

        option_price = \
            gpu_pool.map_reduce(f=simulate_and_compute_option_price,
                                reduction=lambda x, y: x + y / number_of_batches,
                                initial_value=0.0,
                                kwargs_list=number_of_batches * [kwargs])

    return option_price



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [(np.array([1, 5, 3, 2],
                       dtype=np.int32),
              np.array([50, 200, -5, 10],
                       dtype=np.int32))]


@pytest.mark.parametrize("keys_numpy, values_numpy", test_data)
def test_sort_by_keys(keys_numpy, values_numpy):
    cocos.device.init()

    keys_cocos = cn.array(keys_numpy)
    values_cocos = cn.array(values_numpy)

    # out_keys_numpy, out_values_numpy = np.sort_by_keys(keys_numpy, values_numpy, axis=0)
    # out_keys_cocos, out_values_cocos = cn.sort_by_keys(keys_cocos, values_cocos, axis=0)
    #
    # assert np.allclose(out_keys_cocos, out_keys_numpy)
    # assert np.allclose(out_values_cocos, out_values_numpy)



import cocos.device
import cocos.numerics as cn
import cocos.numerics.linalg
import numpy as np
import pytest

test_data = [np.inf, 1, 2, None]


@pytest.mark.parametrize("ord", test_data)
def test_norm(ord):
    cocos.device.init()

    # using numpy
    A_numpy = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 20]], dtype=np.float32)
    print(A_numpy)

    norm = np.linalg.norm(A_numpy, ord)
    print(f"norm python = {norm}")

    # using Cocos
    A_cocos = cn.array(A_numpy)

    norm_cocos = cn.linalg.norm(A_cocos, ord)
    print(f"norm cocos = {norm_cocos}")

    # conduct tests
    assert np.isclose(norm, norm_cocos)



import pytest

import cocos.numerics as cn
from cocos.tests.test_numerics.test_statistics.utilities import perform_ks_test


n_kolmogorov_smirnov = 10000
test_data = [(0.5, 0.5, n_kolmogorov_smirnov),
             (5.0, 1.0, n_kolmogorov_smirnov),
             (1.0, 3.0, n_kolmogorov_smirnov),
             (2.0, 2.0, n_kolmogorov_smirnov),
             (2.0, 5.0, n_kolmogorov_smirnov)]


@pytest.mark.parametrize("a, b, n_kolmogorov_smirnov", test_data)
def test_beta_distribution(a, b, n_kolmogorov_smirnov):
    u = cn.random.beta(a, b, n_kolmogorov_smirnov)
    print(u.shape)
    reject = perform_ks_test(u,
                             alpha=0.01,
                             distribution='beta',
                             args=(a, b),
                             verbose=True)

    assert not reject



from arrayfire.library import backend, c_char_ptr_t
from arrayfire.util import safe_call, to_str, c_pointer
import math
import typing as tp


def _pad_shape_tuple_default(shape: tp.Union[tp.Sequence, int, float],
                             d0: tp.Optional[int],
                             d1: tp.Optional[int],
                             d2: tp.Optional[int],
                             d3: tp.Optional[int]) \
        -> tp.Tuple[int,
                    tp.Optional[int],
                    tp.Optional[int],
                    tp.Optional[int]]:
    if isinstance(shape, list) or isinstance(shape, int):
        shape = tuple(shape)

    ndim = len(shape)
    if ndim < 1 or ndim > 4:
        raise ValueError("the array must have between 1 and 4 axes")
    else:
        d0 = shape[0]
        if ndim > 1:
            d1 = shape[1]
            if ndim > 2:
                d2 = shape[2]
                if ndim > 3:
                    d3 = shape[3]

    return d0, d1, d2, d3


def _pad_shape_tuple_none(shape: tp.Union[tp.Sequence, int, float]) \
        -> tp.Tuple[int,
                    tp.Optional[int],
                    tp.Optional[int],
                    tp.Optional[int]]:
    return _pad_shape_tuple_default(shape=shape,
                                    d0=None,
                                    d1=None,
                                    d2=None,
                                    d3=None)


def _pad_shape_tuple_axis(axes: tp.Sequence[int]) \
        -> tp.Tuple[int, int, int, int]:
    output_shape = list(axes)
    for i in range(len(axes), 4):
        output_shape.append(i)

    return tuple(output_shape)


def _pad_shape_tuple_one(shape: tp.Tuple[int]) -> tp.Tuple[int, int, int, int]:
    d0 = 1
    d1 = 1
    d2 = 1
    d3 = 1
    ndim = len(shape)
    if ndim < 1 or ndim > 4:
        raise ValueError("the array must have between 1 and 4 axes")
    else:
        d0 = shape[0]
        if ndim > 1:
            d1 = shape[1]
            if ndim > 2:
                d2 = shape[2]
                if ndim > 3:
                    d3 = shape[3]

    return d0, d1, d2, d3


def _as_str(self, dims: bool = True):
    arr_str = c_char_ptr_t(0)
    be = backend.get()
    safe_call(be.af_array_to_string(c_pointer(arr_str), "", self.arr, 4, dims))
    py_str = to_str(arr_str)
    safe_call(be.af_free_host(arr_str))
    return py_str


def _compute_slice_length(index: slice, length: int) -> int:
    """
    Given a Python slice such as 1:8:2 and the length of a source sequence
    (e.g. a numpy array) this function computes the resulting number of elements
    when using the slice to index the source sequence.

    :param index: a Python slice
    :param length: length of the source sequence
    :return: the number of elements resulting from the indexing
    """

    start, stop, step = index.indices(length)
    # print(f'start = {start}, stop = {stop}, step = {step}')
    computed_length = math.floor((stop - start - 1) / step) + 1
    # print(f'computed length = {computed_length}')
    return computed_length


def _remove_trailing_ones(shape: tp.Tuple[int, ...]) -> tp.Tuple[int, ...]:
    while len(shape) > 1 and shape[-1] == 1:
        shape = shape[:-1]

    return shape



from contexttimer import Timer
import math
import matplotlib.pyplot as plt
import multiprocessing
import numpy
import pandas as pd
import time
import typing as tp

import cocos.device
from cocos.multi_processing.device_pool import ComputeDevicePool
from cocos.numerics.random import randn_antithetic
from cocos.numerics.numerical_package_bundle import (
    get_available_numerical_packages,
    NumericalPackageBundle,
    NumpyBundle,
    CocosBundle
)

from examples.stochastic_volatility.heston_utilities import (
    simulate_and_compute_option_price_gpu,
    simulate_and_compute_option_price_multicore,
    simulate_and_compute_option_price,
    simulate_heston_model,
    compute_option_price_from_simulated_paths
)


def loky_installed() -> bool:
    try:
        import loky
        return True
    except:
        return False


class NumpyMulticoreBundle(NumpyBundle):
    @classmethod
    def is_installed(cls) -> bool:
        return super().is_installed() & loky_installed()

    @classmethod
    def label(cls) -> str:
        return 'NumPy Multicore'


class CocosMultiGPUBundle(CocosBundle):
    def __init__(self, number_of_gpus: int):
        self.number_of_gpus = number_of_gpus

    @classmethod
    def is_installed(cls) -> bool:
        return super().is_installed() & loky_installed()

    def label(self) -> str:
        if self.number_of_gpus == 1:
            return 'Cocos on 1 GPU'
        else:
            return f'Cocos on {self.number_of_gpus} GPUs'


class HestonBenchmarkResults:
    def __init__(self,
                 numerical_package_bundle: tp.Type[NumericalPackageBundle],
                 total_time: float,
                 option_price: float):
        self._numerical_package_bundle = numerical_package_bundle
        self._total_time = total_time
        self._option_price = option_price

    @property
    def numerical_package_bundle(self) -> tp.Type[NumericalPackageBundle]:
        return self._numerical_package_bundle

    @property
    def total_time(self) -> float:
        return self._total_time

    @property
    def option_price(self) -> float:
        return self._option_price

    def print_results(self):
        label = self.numerical_package_bundle.label()
        print(f'results on {label}')
        print(
            f"Time using {label}: {self.total_time} secs")
        print(f"Call option price using {label}: {self.option_price}")


def run_benchmark(x0: float,
                  v0: float,
                  r: float,
                  rho: float,
                  sigma_v: float,
                  kappa: float,
                  v_bar: float,
                  T: float,
                  K: float,
                  nT: int,
                  R: int,
                  numerical_package_bundle: tp.Type[NumericalPackageBundle]) \
        -> HestonBenchmarkResults:

    # number of paths for warm-up (compile GPU kernels)
    R_warm_up = 1000
    _, _ = \
        simulate_heston_model(T=T,
                              N=nT,
                              R=R_warm_up,
                              mu=r,
                              kappa=kappa,
                              v_bar=v_bar,
                              sigma_v=sigma_v,
                              rho=rho,
                              x0=x0,
                              v0=v0,
                              numerical_package_bundle=numerical_package_bundle)

    # actual simulation run to price plain vanilla call option
    with Timer() as timer:
        (x_simulated, v_simulated) \
            = simulate_heston_model(
                T=T,
                N=nT,
                R=R,
                mu=r,
                kappa=kappa,
                v_bar=v_bar,
                sigma_v=sigma_v,
                rho=rho,
                x0=x0,
                v0=v0,
                numerical_package_bundle=numerical_package_bundle)

        # compute option price
        option_price \
            = compute_option_price_from_simulated_paths(
                r=r,
                T=T,
                K=K,
                x_simulated=x_simulated,
                numerical_package_bundle=numerical_package_bundle)

    # print(option_price)
    numerical_package_bundle.synchronize()
    total_time = timer.elapsed

    return HestonBenchmarkResults(numerical_package_bundle,
                                  total_time=total_time,
                                  option_price=option_price)


def run_benchmarks(
        numerical_package_bundles:
        tp.Optional[tp.Tuple[tp.Type[NumericalPackageBundle], ...]] = None) \
        -> tp.Dict[type(NumericalPackageBundle), HestonBenchmarkResults]:
    if numerical_package_bundles is None:
        numerical_package_bundles = \
            get_available_numerical_packages(list_installed_bundles=True)

    # model parameters
    x0 = 0.0  # initial log stock price
    v0 = 0.101 ** 2  # initial volatility
    r = math.log(1.0319)  # risk-free rate
    rho = -0.7  # instantaneous correlation between Brownian motions
    sigma_v = 0.61  # variance of volatility
    kappa = 6.21  # mean reversion speed
    v_bar = 0.019  # mean variance

    # option parameters
    T = 1.0  # time to expiration
    K = 0.95  # strike price

    # simulation parameters
    nT = int(math.ceil(500 * T))  # number of time-steps to simulate
    R = 2000000  # actual number of paths to simulate for pricing

    kwargs = \
        dict(x0=x0,
             v0=v0,
             r=r,
             rho=rho,
             sigma_v=sigma_v,
             kappa=kappa,
             v_bar=v_bar,
             T=T,
             K=K,
             nT=nT,
             R=R)

    # single core benchmarks
    numerical_package_bundle_to_result_map = {}
    for numerical_package_bundle in numerical_package_bundles:
        kwargs['numerical_package_bundle'] = numerical_package_bundle
        # run benchmark on gpu
        heston_benchmark_results \
            = run_benchmark(**kwargs)

        numerical_package_bundle_to_result_map[numerical_package_bundle] \
            = heston_benchmark_results

        heston_benchmark_results.print_results()

    if loky_installed():
        # multi core benchmarks
        kwargs['numerical_package_bundle'] = NumpyMulticoreBundle
        # initialize Python processes
        with Timer() as timer:
            _ \
                = simulate_and_compute_option_price_multicore(**kwargs)
            print(f'time in first run={timer.elapsed}')

        with Timer() as timer:
            option_price \
                = simulate_and_compute_option_price_multicore(**kwargs)

        total_time = timer.elapsed

        numpy_multicore_results = \
            HestonBenchmarkResults(
                numerical_package_bundle=NumpyMulticoreBundle,
                total_time=total_time,
                option_price=option_price)

        numpy_multicore_results.print_results()

        numerical_package_bundle_to_result_map[NumpyMulticoreBundle] \
            = numpy_multicore_results

        # multi gpu benchmarks
        if 'numerical_package_bundle' in kwargs:
            kwargs.pop('numerical_package_bundle')

        gpu_pool = ComputeDevicePool()
        if gpu_pool.number_of_devices > 1:
            # warm up
            kwargs['R'] = 20000
            option_price = \
                simulate_and_compute_option_price_gpu(gpu_pool=gpu_pool,
                                                      **kwargs)

            # actual benchmark
            kwargs['R'] = R

            for number_of_gpus in range(1, gpu_pool.number_of_devices + 1):
                cocos_multi_gpu_bundle = CocosMultiGPUBundle(number_of_gpus=number_of_gpus)

                with Timer() as timer:
                    option_price = \
                        simulate_and_compute_option_price_gpu(gpu_pool=gpu_pool,
                                                              number_of_batches=number_of_gpus,
                                                              **kwargs)
                    cocos.device.sync()

                total_time = timer.elapsed

                cocos_multi_gpu_results = \
                    HestonBenchmarkResults(
                        numerical_package_bundle=cocos_multi_gpu_bundle,
                        total_time=total_time,
                        option_price=option_price)

                cocos_multi_gpu_results.print_results()

                numerical_package_bundle_to_result_map[cocos_multi_gpu_bundle] \
                    = cocos_multi_gpu_results
    else:
        print(f'Please install loky to enable multi core and multi gpu benchmarks')

    return numerical_package_bundle_to_result_map


def create_result_dataframe(
        numerical_package_bundle_to_result_map:
        tp.Dict[type(NumericalPackageBundle), HestonBenchmarkResults]) \
        -> pd.DataFrame:
    numpy_results = numerical_package_bundle_to_result_map[NumpyBundle]

    table_dict = {}

    for numerical_package_bundle, result \
            in numerical_package_bundle_to_result_map.items():
        entry_dict = {'Total Time in Seconds': result.total_time,
                      'Speedup Compared to NumPy':
                          numpy_results.total_time / result.total_time}
        table_dict[numerical_package_bundle.label()] = entry_dict

    return pd.DataFrame(table_dict).transpose()


def create_result_table(
        numerical_package_bundle_to_result_map:
        tp.Dict[type(NumericalPackageBundle), HestonBenchmarkResults]) -> str:
    numpy_results = numerical_package_bundle_to_result_map[NumpyBundle]

    res = "<table>\n"
    res += "<tbody>\n"
    res += "<tr>\n"
    res += "<th></th>\n"
    res += "<th>Total Time in Seconds</th>\n"
    res += "<th>Speedup Compared to NumPy</th>\n"
    res += "</tr>\n"

    for numerical_package_bundle, result \
            in numerical_package_bundle_to_result_map.items():
        res += "<tr>\n"
        res += f"<td>{numerical_package_bundle.label()}</td>\n"
        res += f"<td>{result.total_time}</td>\n"
        res += f"<td>{numpy_results.total_time / result.total_time}</td>\n"
        res += "</tr>\n"

    res += "</table>"

    return res


def create_bar_plot(
        numerical_package_bundle_to_result_map:
        tp.Dict[type(NumericalPackageBundle), HestonBenchmarkResults]):
    numpy_results = numerical_package_bundle_to_result_map[NumpyBundle]

    objects = [numerical_package_bundle.label()
               for numerical_package_bundle
               in numerical_package_bundle_to_result_map.keys()]

    y_pos = numpy.arange(len(objects))

    performance = [numpy_results.total_time / result.total_time
                   for result
                   in numerical_package_bundle_to_result_map.values()]

    plt.figure(1)
    plt.bar(y_pos, performance, align='center', alpha=0.5)
    plt.xticks(y_pos, objects, rotation=45, ha="right")
    plt.ylabel('Speedup Factor')
    plt.title('Performance Relative to NumPy \n'
              'in Monte Carlo Simulation of Heston Model \n')

    plt.savefig(f'heston_pricing_benchmark_results', bbox_inches='tight')

    plt.show()


def main():
    numerical_package_bundle_to_result_map = run_benchmarks()

    # create result dataframe
    result_df = create_result_dataframe(numerical_package_bundle_to_result_map)
    print(result_df)

    # create result table
    result_table = create_result_table(numerical_package_bundle_to_result_map)
    print(result_table)

    # plot and save result bar graph
    create_bar_plot(numerical_package_bundle_to_result_map)


if __name__ == '__main__':
    main()



import numpy as np
import typing as tp

import cocos.numerics as cn
from cocos.numerics.data_types import NumericArray


def find_length_of_state_vectors(state_vectors: tp.Iterable[NumericArray]) \
        -> int:
    R = max([state_argument.size
         for state_argument
         in state_vectors
         if isinstance(state_argument, (np.ndarray, cn.ndarray))] + [1])

    return R



import cocos.device
import cocos.numerics as cn
import cocos.numerics.linalg as linalg
import numpy as np

# backend = 'cpu'
# backend = 'cuda'
# backend = 'opencl'
backend = None


def compare_cocos_numpy(cocos_array, numpy_array):
    return np.allclose(np.array(cocos_array), numpy_array)


def test_svd():
    cocos.device.init(backend)
    cocos.device.info()

    # selected_backend = device_info['backend'].lower()
    selected_backend = cocos.device.selected_backend()
    # print("backend lowercase = <{}>".format(selected_backend))
    # print("device = {}".format(device_info['device']))
    # print("backend = {}".format(device_info['backend']))
    # print("toolkit = {}".format(device_info['toolkit']))
    # print("compute = {}".format(device_info['compute']))

    dtype = np.float32
    if cocos.device.is_dbl_supported():
        dtype = np.float64

    A = np.array([[3, 2, 2], [2, 3, -2]], dtype=dtype)

    # numpy
    U, s, V = np.linalg.svd(A)
    # print("U numpy")
    # print(U)
    # print("")
    #
    # print("s")
    # print(s)
    # print("")
    #
    # print("V")
    # print(V)
    # print("")

    S = np.zeros((2, 3), dtype=dtype)
    S[:2, :2] = np.diag(s)
    reconstructed_A = np.dot(U, np.dot(S, V))
    assert np.allclose(reconstructed_A, A)

    # Cocos
    A_cocos = cn.array(A)
    U_cocos, s_cocos, V_cocos = cn.linalg.svd(A_cocos)
    print("U cocos")
    print(np.array(U_cocos))

    print("s cocos")
    print(np.array(s_cocos))

    print("V cocos")
    print(np.array(V_cocos))

    # S_cocos = cocos.diag(s_cocos)
    # reconstructed_A_cocos = cocos.dot(cocos.dot(U_cocos, S_cocos), V_cocos)

    S_cocos = cn.zeros((2, 3), dtype=dtype)
    # print("diagonalized cocos s array")
    # print(cocos.diag(s_cocos))
    # print("diag(s_cocos).dtype = {}".format(cocos.diag(s_cocos).dtype))
    S_cocos[0:2, 0:2] = cn.diag(s_cocos)
    first_step = cn.dot(S_cocos, V_cocos)
    second_step = cn.dot(U_cocos, first_step)
    reconstructed_A_cocos = second_step
    assert np.allclose(reconstructed_A_cocos, A_cocos)
    assert compare_cocos_numpy(s_cocos, s)

    if cocos.device.is_dbl_supported() and selected_backend != 'cuda' and selected_backend != 'opencl':
        assert compare_cocos_numpy(U_cocos, U)
        assert compare_cocos_numpy(V_cocos, V)



import typing as tp

from cocos.multi_processing.utilities import (
    ResultType,
    MultiprocessingPoolType,
    _extract_arguments_and_number_of_batches
)


def map_reduce_multicore(
        f: tp.Callable[..., ResultType],
        reduction: tp.Callable[[ResultType, ResultType], ResultType],
        initial_value: ResultType,
        args_list: tp.Optional[tp.Sequence[tp.Sequence]] = None,
        kwargs_list: tp.Optional[tp.Sequence[tp.Dict[str, tp.Any]]] = None,
        number_of_batches: tp.Optional[int] = None,
        multiprocessing_pool_type: MultiprocessingPoolType = MultiprocessingPoolType.default()) \
        -> ResultType:
    """
    This function evaluates the function 'f' on elements of 'args_list' and
    'kwargs_list' in parallel on multiple cpu cores and performs the reduction
    by calling the function 'reduction' on the result and the result of the
    reductions so far to eventually produce one final result of type
    'ResultType'. The reduce step is performed from the left and results are
    being processed in the same order as they appear in `args_list` and
    `kwargs_list`.

    If the arguments for each run of 'f' are identical and they have already
    been applied to the function that is passed then 'args_list' and
    'kwargs_list' may both be None but the argument 'number_of_batches' must
    be specified so the method knows how many times to run the function 'f'.

    Args:
        f: The map function to be evaluated over elements of 'args_list' and
               'kwargs_list'.

        reduction: The reduction to be performed on the results of 'f'.
                       This is done on the host (not the device).

        initial_value: The initial value of the reduction
                           (i.e. the neutral element).

        args_list: A sequence of sequences of positional arguments.
        kwargs_list: A sequence of dictionaries of keyword arguments.
        number_of_batches:
            The number of function evaluations is required if 'args_list'
            and 'kwargs_list' are both empty.

        multiprocessing_pool_type:
            the type of multi-processing pool (see class MultiprocessingPoolType)
    """
    args_list, kwargs_list, number_of_batches = \
        _extract_arguments_and_number_of_batches(
            args_list=args_list,
            kwargs_list=kwargs_list,
            number_of_batches=number_of_batches)

    def wrapped_f(index, *args, **kwargs) -> ResultType:
        return index, f(*args, **kwargs)

    if multiprocessing_pool_type == MultiprocessingPoolType.LOKY:
        from concurrent.futures import as_completed
        from loky import get_reusable_executor

        executor = \
            get_reusable_executor(timeout=None,
                                  context='loky')

        futures = [executor.submit(wrapped_f, i, *args, **kwargs)
                   for i, (args, kwargs)
                   in enumerate(zip(args_list, kwargs_list))]

        result_from_future = lambda x: x.result()
    elif multiprocessing_pool_type == MultiprocessingPoolType.PATHOS:
        from pathos.pools import ProcessPool
        pool = ProcessPool()

        futures = [pool.apipe(wrapped_f, i, *args, **kwargs)
                   for i, (args, kwargs)
                   in enumerate(zip(args_list, kwargs_list))]

        result_from_future = lambda x: x.get()
    else:
        raise ValueError(f'Multiprocessing pool type {multiprocessing_pool_type} not supported')

    results = [result_from_future(future) for future in futures]
    results = sorted(results, key=lambda x: x[0])
    results = [result[1] for result in results]

    result = initial_value
    for new_result in results:
        result = reduction(result, new_result)

    return result


def map_combine_multicore(
        f: tp.Callable[..., ResultType],
        combination: tp.Callable[[tp.Iterable[ResultType]], ResultType],
        args_list: tp.Optional[tp.Sequence[tp.Sequence]] = None,
        kwargs_list: tp.Optional[tp.Sequence[tp.Dict[str, tp.Any]]] = None,
        number_of_batches: tp.Optional[int] = None,
        multiprocessing_pool_type: MultiprocessingPoolType = MultiprocessingPoolType.default()) \
        -> ResultType:
    """
    This function evaluates the function `f` on elements of `args_list` and
    `kwargs_list` in parallel on multiple cpu cores and aggregates results
    in a single step by calling the function `combination` with a list of all
    results. Results provided to `combination` are in the same order as
    they appear in `args_list` and `kwargs_list`.

    If the arguments for each run of 'f' are identical and they have already
    been applied to the function that is passed then 'args_list' and
    'kwargs_list' may both be None but the argument 'number_of_batches' must
    be specified so the method knows how many times to run the function 'f'.

    Args:
        f: The map function to be evaluated over elements of 'args_list' and
               'kwargs_list'.

        combination: A function that aggregates a list of all results in a single step
        args_list: A sequence of sequences of positional arguments.
        kwargs_list: A sequence of dictionaries of keyword arguments.
        number_of_batches:
            The number of function evaluations is required if 'args_list'
            and 'kwargs_list' are both empty.

        multiprocessing_pool_type:
            the type of multi-processing pool (see class MultiprocessingPoolType)
    """
    args_list, kwargs_list, number_of_batches = \
        _extract_arguments_and_number_of_batches(
            args_list=args_list,
            kwargs_list=kwargs_list,
            number_of_batches=number_of_batches)

    def wrapped_f(index, *args, **kwargs) -> ResultType:
        return index, f(*args, **kwargs)

    results = []
    if multiprocessing_pool_type == MultiprocessingPoolType.LOKY:
        from concurrent.futures import as_completed
        from loky import get_reusable_executor

        executor = \
            get_reusable_executor(timeout=None,
                                  context='loky')

        futures = [executor.submit(wrapped_f, i, *args, **kwargs)
                   for i, (args, kwargs)
                   in enumerate(zip(args_list, kwargs_list))]

        for future in as_completed(futures):
            results.append(future.result())
    elif multiprocessing_pool_type == MultiprocessingPoolType.PATHOS:
        from pathos.pools import ProcessPool
        pool = ProcessPool()

        futures = [pool.apipe(wrapped_f, i, *args, **kwargs)
                   for i, (args, kwargs)
                   in enumerate(zip(args_list, kwargs_list))]

        for future in futures:
            results.append(future.get())
    else:
        raise ValueError(f'Multiprocessing pool type {multiprocessing_pool_type} not supported')

    results = sorted(results, key=lambda x: x[0])
    results = [result[1] for result in results]

    # print(f'len(results)={len(results)}')
    # for result in results:
    #     print(result.shape)

    return combination(results)


# class MultiCoreMapReduce:
#     def __init__(self, number_of_cores: tp.Optional[int] = None):
#         self._number_of_cores = number_of_cores



import numpy as np
import pytest

from cocos.numerics._utilities import _compute_slice_length


test_data = [(slice(0, 3, 2), np.array([1, 2, 3, 4, 5, 6, 7, 8])),
             (slice(1, 3, 2), np.array([1, 2, 3, 4, 5, 6, 7, 8])),
             (slice(2, 3, 2), np.array([1, 2, 3, 4, 5, 6, 7, 8])),
             (slice(3, 3, 2), np.array([1, 2, 3, 4, 5, 6, 7, 8])),
             (slice(1, 3, 3), np.array([1, 2, 3, 4, 5, 6, 7, 8])),
             (slice(1, 2, 2), np.array([1, 2, 3, 4, 5, 6, 7, 8])),
             (slice(3, 3, 2), np.array([1, 2, 3, 4, 5, 6, 7, 8]))]


@pytest.mark.parametrize("index, x", test_data)
def test_compute_slice_length(index: slice, x: np.ndarray):
    y = x[index]
    computed_length = _compute_slice_length(index, len(x))
    if not len(y) == computed_length:
        print(f'computed length = {computed_length}, actual length = {len(y)}')
    assert len(y) == computed_length



import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1, -2, 5, 6, 20, 50, 3, 6],
                       [-3, 4, 7, 8, 10, 60, 70, 100],
                       [-30, 40, 70, 80, 100, 600, 700, 1000],
                       [9, 3, 1, 27, 81, 84, 87, 78],
                       [90, 30, 10, 270, 810, 840, 870, 780]],
                      dtype=np.int32)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_sort(A_numpy):
    cn.init()

    A_cocos = cn.array(A_numpy)

    for i in range(2):
        # argsort with axis
        argsort_numpy = np.argsort(A_numpy, axis=i)
        argsort_cocos = cn.argsort(A_cocos, axis=i)
        assert np.allclose(argsort_cocos, argsort_numpy)

        # sort with axis
        sort_numpy = np.sort(A_numpy, axis=i)
        sort_cocos = cn.sort(A_cocos, axis=i)
        assert np.allclose(sort_cocos, sort_numpy)

    # argsort without axis
    argsort_numpy = np.argsort(A_numpy, axis=None)
    argsort_cocos = cn.argsort(A_cocos.transpose(), axis=None)
    # print("argsort numpy")
    # print(argsort_numpy)
    # print("argsort cocos")
    # print(argsort_cocos)
    # assert np.allclose(argsort_cocos, argsort_numpy)

    # sort without axis
    sort_numpy = np.sort(A_numpy)
    sort_cocos = cn.sort(A_cocos)
    assert np.allclose(sort_cocos, sort_numpy)



import cocos.device
import cocos.numerics as cn
import cocos.numerics.linalg as cnl
import numpy as np
import pytest

test_data = [np.array([[1, 2, 3], [4, 5, 6], [7, 8, 20]],
                      dtype=np.float32),
             np.array([[1 + 2j, 2, 3], [4, 5 + 1j, 6], [7, 8, 20 + 7j]],
                      dtype=np.complex64)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_det_inv_rank(A_numpy):
    cocos.device.init()

    # using numpy
    # print(A_numpy)

    det_numpy = np.linalg.det(A_numpy)
    rank_numpy = np.linalg.matrix_rank(A_numpy)
    inv_numpy = np.linalg.inv(A_numpy)

    print(f"determinant python = {det_numpy}")

    # using Cocos
    A_cocos = cn.array(A_numpy)

    det_cocos = cnl.det(A_cocos)
    rank_cocos = cnl.matrix_rank(A_cocos)
    inv_cocos = cnl.inv(A_cocos)
    print(f"determinant cocos= {det_cocos}")

    # conduct tests
    assert np.isclose(det_numpy, det_cocos)
    assert np.isclose(rank_numpy, rank_cocos)
    assert np.allclose(inv_numpy, inv_cocos)


import cocos.device
import cocos.numerics as cn
import numpy as np


def test_union_intersection():
    cocos.device.init()

    A_numpy = np.array([1, 2, 3], dtype=np.int32)
    B_numpy = np.array([1, 5], dtype=np.int32)

    A_cocos = cn.array(A_numpy)
    B_cocos = cn.array(B_numpy)

    union_AB_numpy = np.union1d(A_numpy, B_numpy)
    intersection_AB_numpy = np.union1d(A_numpy, B_numpy)

    union_AB_cocos = cn.union1d(A_cocos, B_cocos)
    intersection_AB_cocos = cn.union1d(A_cocos, B_cocos)

    assert np.allclose(union_AB_cocos, union_AB_numpy)
    assert np.allclose(intersection_AB_cocos, intersection_AB_numpy)




import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest

test_data = [(np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 20]],
                       dtype=np.int32),
              (2, 3)),
             (np.array([[0.2, 1.0, 0.5],
                        [0.4, 0.5, 0.6],
                        [0.7, 0.2, 0.25]],
                       dtype=np.float32),
              (2, 3)),
             (np.array([[0.5, 2.3, 3.1],
                        [4, 5.5, 6],
                        [7 - 9j, 8 + 1j, 2 + 10j]],
                       dtype=np.complex64),
              (2, 3)),
             (np.array([[[1.0, 2], [3, 4]],
                        [[5, 6], [7, 8]]],
                       dtype=np.float32),
              (2, 3, 1)),
             (np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 20]],
                       dtype=np.int32), (2, 1)),
             (np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 20]],
                       dtype=np.int32),
              (1, 3))
            ]


@pytest.mark.parametrize("A_numpy, tiles", test_data)
def test_tile(A_numpy, tiles):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    B_numpy = np.tile(A_numpy, tiles)
    B_cocos = cn.tile(A_cocos, tiles)

    assert np.allclose(B_cocos, B_numpy)



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1, -1],
                       [0, 1]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.2, 0.25]],
                      dtype=np.float32)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_trigonometric(A_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    all_positive = np.all(A_numpy > 0)

    all_finite = cn.all(cn.isfinite(A_cocos))
    assert np.allclose(np.isfinite(A_numpy), cn.isfinite(A_cocos))

    if all_finite:
        assert np.allclose(np.sinh(A_numpy), cn.sinh(A_cocos))
        assert np.allclose(np.cosh(A_numpy), cn.cosh(A_cocos))
        assert np.allclose(np.tanh(A_numpy), cn.tanh(A_cocos))
        assert np.allclose(np.sin(A_numpy), cn.sin(A_cocos))
        assert np.allclose(np.cos(A_numpy), cn.cos(A_cocos))
        assert np.allclose(np.tan(A_numpy), cn.tan(A_cocos))

        assert np.allclose(np.arcsin(np.sin(A_numpy)),
                           cn.arcsin(cn.sin(A_cocos)))

        assert np.allclose(np.arccos(np.cos(A_numpy)),
                           cn.arccos(cn.cos(A_cocos)))

        assert np.allclose(np.arctan(np.tan(A_numpy)),
                           cn.arctan(cn.tan(A_cocos)))

        assert np.allclose(np.arcsinh(np.sinh(A_numpy)),
                           cn.arcsinh(cn.sinh(A_cocos)))

        assert np.allclose(np.arccosh(np.cosh(A_numpy)),
                           cn.arccosh(cn.cosh(A_cocos)))

        assert np.allclose(np.arctanh(np.tanh(A_numpy)),
                           cn.arctanh(cn.tanh(A_cocos)))

        if cn.isfloating(A_cocos):
            assert np.allclose(A_cocos, cn.arcsin(cn.sin(A_cocos)))
            assert np.allclose(A_cocos, cn.arccos(cn.cos(A_cocos)))
            assert np.allclose(A_cocos, cn.arctan(cn.tan(A_cocos)))
            assert np.allclose(A_cocos, cn.arcsinh(cn.sinh(A_cocos)))
            assert np.allclose(A_cocos, cn.arccosh(cn.cosh(A_cocos)))
            assert np.allclose(A_cocos, cn.arctanh(cn.tanh(A_cocos)))



import pytest

import cocos.numerics as cn
from cocos.tests.test_numerics.test_statistics.utilities import perform_ks_test


n_kolmogorov_smirnov = 1000000
test_data = [(0, 1, n_kolmogorov_smirnov),
             (2, 5, n_kolmogorov_smirnov),
             (3, 12, n_kolmogorov_smirnov),
             (5, 10, n_kolmogorov_smirnov),
             (1, 2, n_kolmogorov_smirnov),
             (10, 50, n_kolmogorov_smirnov),
             (20, 25, n_kolmogorov_smirnov)]


@pytest.mark.parametrize("a, b, n_kolmogorov_smirnov", test_data)
def test_uniform_distribution(a, b, n_kolmogorov_smirnov):
    u = cn.random.uniform(a, b, n_kolmogorov_smirnov)
    reject = perform_ks_test(u,
                             alpha=0.01,
                             distribution='uniform',
                             args=(a, b - a))

    assert not reject



import pytest

import cocos.numerics as cn
from cocos.tests.test_numerics.test_statistics.utilities import perform_ks_test


n_kolmogorov_smirnov = 2000000
test_data = [(1, n_kolmogorov_smirnov),
             (2, n_kolmogorov_smirnov),
             (3, n_kolmogorov_smirnov),
             (4, n_kolmogorov_smirnov),
             (5, n_kolmogorov_smirnov),
             (6, n_kolmogorov_smirnov),
             (9, n_kolmogorov_smirnov)]


@pytest.mark.parametrize("df, n_kolmogorov_smirnov", test_data)
def test_chisquare_distribution(df, n_kolmogorov_smirnov):
    u = cn.random.chisquare(df, n_kolmogorov_smirnov)
    reject = perform_ks_test(u,
                             alpha=0.01,
                             distribution='chi2',
                             args=(df, 0.0, 0.0),
                             verbose=True)

    assert not reject



import math
import matplotlib.pyplot as plt
import numpy
import time
import typing as tp

from cocos.device import (
    ComputeDeviceManager,
    info,
    sync
)

from cocos.multi_processing.device_pool import ComputeDevicePool
from cocos.multi_processing.multi_core_batch_processing import map_reduce_multicore

from cocos.numerics.numerical_package_bundle import (
    NumericalPackageBundle,
    CocosBundle
)

from cocos.numerics.random import randn_antithetic

from stochastic_volatility.heston_utilities import (
    simulate_and_compute_option_price_gpu,
    simulate_and_compute_option_price_multicore,
    simulate_and_compute_option_price,
    simulate_heston_model,
    compute_option_price_from_simulated_paths
)


def create_result_table(number_of_devices_to_runtime_map: tp.Dict[int, float]) \
        -> str:
    res = "<table>\n"
    res += "<tbody>\n"
    res += "<tr>\n"
    res += "<th>Number of GPUs</th>\n"
    res += "<th>Total Time in Seconds</th>\n"
    res += "<th>Speedup Compared to Single GPU</th>\n"
    res += "</tr>\n"

    for number_of_devices, runtime \
            in number_of_devices_to_runtime_map.items():
        res += "<tr>\n"
        res += f"<td>{number_of_devices}</td>\n"
        res += f"<td>{runtime}</td>\n"
        res += f"<td>{number_of_devices_to_runtime_map[1] / runtime}</td>\n"
        res += "</tr>\n"

    res += "</table>"

    return res


def create_bar_plot(number_of_devices_to_runtime_map: tp.Dict[int, float]):
    objects = []
    performance = []

    for number_of_devices, runtime \
            in number_of_devices_to_runtime_map.items():
        objects.append(number_of_devices)
        performance.append(number_of_devices_to_runtime_map[1] / runtime)

    y_pos = numpy.arange(len(objects))

    plt.figure(1)
    plt.bar(y_pos, performance, align='center', alpha=0.5)
    plt.xticks(y_pos, objects)
    plt.ylabel('Speedup Factor')
    plt.title('Performance Relative to a Single GPU \n'
              'in Monte Carlo Simulation of Heston Model \n')

    plt.savefig(f'heston_pricing_benchmark_results_multi_gpu')

    plt.show()


if __name__ == '__main__':
    info()

    gpu_pool = ComputeDevicePool()

    # model parameters
    x0 = 0.0  # initial log stock price
    v0 = 0.101 ** 2  # initial volatility
    r = math.log(1.0319)  # risk-free rate
    rho = -0.7  # instantaneous correlation between Brownian motions
    sigma_v = 0.61  # variance of volatility
    kappa = 6.21  # mean reversion speed
    v_bar = 0.019  # mean variance

    # option parameters
    T = 1.0  # time to expiration
    K = 0.95  # strike price

    # simulation parameters
    nT = int(math.ceil(500 * T))  # number of time-steps to simulate

    # warm-up
    R = 20000  # actual number of paths to simulate for pricing

    kwargs = \
        dict(x0=x0,
             v0=v0,
             r=r,
             rho=rho,
             sigma_v=sigma_v,
             kappa=kappa,
             v_bar=v_bar,
             T=T,
             K=K,
             nT=nT,
             R=R)

    print('warm-up')
    # tic = time.time()
    option_price = simulate_and_compute_option_price_gpu(gpu_pool=gpu_pool,
                                                         **kwargs)
    # toc = time.time() - tic
    # print(f'option price = {option_price} computed in {toc} seconds')
    print('warmed up')

    # actual run
    R = 2000000  # actual number of paths to simulate for pricing

    kwargs = \
        dict(x0=x0,
             v0=v0,
             r=r,
             rho=rho,
             sigma_v=sigma_v,
             kappa=kappa,
             v_bar=v_bar,
             T=T,
             K=K,
             nT=nT,
             R=R)

    number_of_devices_to_runtime_map = {}

    for i in range(1, gpu_pool.number_of_devices + 1):
        print(f'computing on {i} GPUs')
        tic = time.time()
        option_price = \
            simulate_and_compute_option_price_gpu(gpu_pool=gpu_pool,
                                                  number_of_batches=i,
                                                  **kwargs)
        sync()
        gpu_time = time.time() - tic
        print(f'option price = {option_price} computed on {i} GPUs in '
              f'{gpu_time} seconds')

        number_of_devices_to_runtime_map[i] = gpu_time

    if gpu_pool.number_of_devices > 1:
        for i in range(2, gpu_pool.number_of_devices + 1):
            print(f'Performance on {i} GPUs increased by a factor of'
                  f' {number_of_devices_to_runtime_map[1] / number_of_devices_to_runtime_map[i]} '
                  f'over a single GPU.')

    result_table = create_result_table(number_of_devices_to_runtime_map)
    print(result_table)

    create_bar_plot(number_of_devices_to_runtime_map)
